from .raw import Raw
from .despike import *
from .rotate_wind3D import rotate_wind
from .correlate_delay import remove_lag
from .spectrum_analysis import Spec
from . import eddyqc
from .eddyqc import *
from .import eddyspec
from .eddyspec import *

__all__ = ["Raw", "despikeVM", "rotate_wind", "remove_lag", "Spec"]
__all__ += eddyqc.__all__
__all__ += eddyspec.__all__


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from .CO2 import center_rolling
from .qcutils import detrend

def two_systems(s1, s2, shift_hz = 100, step_hz = 10, thre1 = 10, thres2 = 10, window_size = 601, ifplot = False):
    # # example:
    # two_systems(df2d['CO2'], df5d['CO2'])
    s1 = s1.copy(); s2 = s2.copy()
    s1d = detrend(s1); s2d = detrend(s2)
    # ===========================================================
    # difference between two series
    mean_val = (s1.mean() + s2.mean()) / 2 
    s_dif = []
    for i in np.arange(-shift_hz, shift_hz + 1, step = step_hz):
        s_dif.append((s1d.shift(i) - s2d) / mean_val * 100)
    s_dif = pd.concat(s_dif, axis = 1).abs().min(axis = 1)
    
    index1 = s_dif[s_dif > thre1].index
    index2 = s_dif[s_dif > s_dif.quantile(0.90)].index
    
    index = index1.union(index2)
    # ===========================================================
    # difference between neighbouring 10 hz
    s1_rm = center_rolling(s1, rolling_window = window_size, method = 'mean').abs()
    s2_rm = center_rolling(s2, rolling_window = window_size, method = 'mean').abs()
    
    s1_rmr = (s1[index] / s1_rm[index] - 1).abs() * 100
    s2_rmr = (s2[index] / s2_rm[index] - 1).abs() * 100
    
    # -----------------------------------------------------------------------------
    # plot indicators
    if ifplot:
        fig, axes = plt.subplots(2, 1, figsize = (10, 6), sharex = True, sharey = True)
        axes = axes.flatten()
        ax = axes[0]
        ax.scatter(s1_rmr.index, s1_rmr)
        ax = axes[1]
        ax.scatter(s2_rmr.index, s2_rmr)
    # -----------------------------------------------------------------------------

    index1 =  s1_rmr[s1_rmr > thres2].index
    index2 =  s2_rmr[s2_rmr > thres2].index
    return index, index1, index2

def user_rolling_qc(s, window_size = 601, method = 'mean'):
    s = s.copy()
    s_rm = center_rolling(s, rolling_window = window_size, method = method).abs()
    s_rmr = (s / s_rm - 1).abs() * 100
    return s_rmr


import numpy as np
import pandas as pd
from datetime import datetime

# ==================================================================================
# Read spectra/cospectra data:
def SplitFFT(fnS, format = '%d/%m/%y %H:%M:%S'):
    df = pd.read_csv(fnS)
    df.columns = df.columns.str.strip()
    datetimes = df['Date'].drop_duplicates().values

    if 'Date' in datetimes:
        datetimes = datetimes[np.where(datetimes != 'Date')]
    datetimes = pd.to_datetime(datetimes, format = format)

    df['Signal'] = df['Signal'].str.strip()
    signals = df['Signal'].drop_duplicates().values
    signals = signals[np.where(signals != 'Signal')]

    freq = [c for c in df.columns if all([i.isnumeric() for i in c.split('.',1)])]

    S = {}
    for sig in signals:
        covar = df.loc[df['Signal'] == sig, 'Co/Var']
        covar.index = datetimes
        S[sig + '_covar'] = covar.astype(float)
        spec = df.loc[df['Signal'] == sig, freq]
        spec.index = datetimes
        S[sig] = spec

    frequency = np.array(freq).astype(float)
    frequency = frequency.reshape(1, -1)

    frequency = np.tile(frequency, (len(datetimes), 1))

    S['Frequency'] = frequency
    
    return S

# ==================================================================================
# Read time-series
def remove_duplicated_index(df, keep = 'last'):
    df = df.copy()
    df = df[~df.index.duplicated(keep=keep)]
    df = df.sort_index()
    return df

def get_date_format(x):
    date_formats = ["%d/%m/%y %H:%M:%S", "%d/%m/%Y %H:%M", "%d/%m/%y %I:%M:%S %p"]
    for date_format in date_formats:
        try:
            d = datetime.strptime(x, date_format)
            # print(date_format)
            break
        except ValueError as e:
            pass
    return d

def SplitCSV(p):    
    csv = pd.read_csv(p, index_col = 0)
    # flux.index = pd.to_datetime(flux.index, format = "%d/%m/%y %H:%M:%S")
    # csv.index = str2dt(csv)
    csv = csv[csv.index != 'Date/Time']
    csv.index = csv.index.map(
        lambda x: get_date_format(x)
    )
    csv = remove_duplicated_index(csv)
    csv = csv.replace(" ", np.nan)
    csv.columns = csv.columns.str.strip()
    csv = csv.astype(float)
    return csv

# ==================================================================================
# Convert number, list, pd.Dataframe, and pd.Series to np.ndarray
def numlist2arr(input):
    if isinstance(input, list):
        return np.array(input)
    elif np.isscalar(input):
        return np.array([input])
    elif isinstance(input, pd.DataFrame) | isinstance(input, pd.Series):
        return input.values
    elif isinstance(input, np.ndarray):
        return input
    else:
        raise(Exception('Wrong type'))


# ==================================================================================
# fetch variable from read-in spectra and timeseries dict 
def fetch(data_dict, name, key_map = {}, transpose = False):
    # key_map: name conversion map from user-defined in EdiRe to fixed spectra analysis functions.
    if name in key_map.keys():
        name = key_map[name]
    data = data_dict[name]
    if isinstance(data, pd.Series) | isinstance(data, pd.DataFrame):
        data = data.values
    elif isinstance(data, list):
        data = np.array(data)
        
    if data.ndim == 1:
        data = data[:, np.newaxis]
    if transpose:
        # data = data.reshape(1, -1)
        data = data.T
    return data


from . import base
from .base import *
from . import toolbox
from .toolbox import *
from . import meteo
from .meteo import *

__all__ = ["base", "toolbox", "meteo"]
__all__ += base.__all__
__all__ += toolbox.__all__


import numpy as np
import pandas as pd
from datetime import datetime, timedelta
from .qcutils import *

def get_minute_std(dfd):
    minutes = dfd.index.map(
        lambda x: x.strftime('%Y-%m-%d %H:%M')
    ).drop_duplicates()

    stds = []
    min_out = []
    for m in minutes:
        df_minute = dfd.loc[m, :]
        if len(df_minute) <= 1: continue

        Ts = df_minute['Ts'].copy()
        Tgmp = df_minute['Tgmp'].copy()
        Ts = Ts - Tgmp

        coefs = np.polyfit(np.arange(len(Ts)), Ts.interpolate().bfill().values, 3)
        ffit = np.polyval(coefs, np.arange(len(Ts)))
        ffit = pd.Series(ffit, index = Ts.index)
        Ts = Ts - ffit

        stds.append(Ts.std())
        min_out.append(m)

    stds = pd.Series(stds, index = min_out, name = 'std')
    stds.index = pd.to_datetime(stds.index, format = '%Y-%m-%d %H:%M')
    
    std_index = iqr_qc(stds, *stats4qc(stds))
    return std_index, stds

def qc_Ts(std_index, ts):
    ano = pd.Series(std_index.copy(), name = 'datetime').to_frame()

    start = ano['datetime'].diff(1).map(
        lambda x: np.abs(x.total_seconds())
    )
    end = ano['datetime'].diff(-1).map(
        lambda x: np.abs(x.total_seconds())
    )

    tol = 5 * 60 # seconds
    end = end[end > tol].index.tolist()
    start = start[start > tol].index.tolist()
    start = [0] + start
    end = end + [len(ano) - 1]

    for s, e in zip(start, end):
        dif = (ano['datetime'][e] - ano['datetime'][s]).total_seconds() / 60
        if dif < 2: continue # minute
        s_dt = ano['datetime'][s]
        e_dt = ano['datetime'][e]
        # ts[s_dt: e_dt] = np.nan
        s_dt = s_dt - timedelta(minutes = 10)
        e_dt = e_dt + timedelta(minutes = 10)
        # ts[iqr_qc(ts[s_dt: e_dt], *stats4qc(ts[ts.index.difference(ts[s_dt:e_dt].index)]))] = np.nan
        ts[s_dt: e_dt] = np.nan
    return ts


import struct
import numpy as np
import pandas as pd
from copy import deepcopy

def int2little(val):
    little_hex = struct.pack('I', val)
    # str_little = ''.join(format(x, '02x') for x in little_hex)
    # return str_little.upper()
    return little_hex

def FP22big(val):
    # big_hex = hex(val).replace('0x', '\\x')
    big_hex = struct.pack('>H', val)
    # # or:
    # str_big = big_hex[2::]
    # return str_big.upper()
    # return big_hex.encode('utf-8')
    return big_hex

def val2FP2(val):
    if np.isnan(val):
        return 0x9ffe
    elif np.isinf(val):
        if val > 0:
            return 0x1fff
        else:
            return 0x9fff
    else:
        if val < 0:

            isNeg = 0x8000
        else:
            isNeg = 0x0
        val = np.abs(val)
        expnt = 4 - len(str(val).split(".")[0])
        mantis = int(val * 10**expnt)

        return mantis | (expnt << 13) | isNeg

def dataframe2tob1(df, p_header, p_save, nlines = 5, lname = 1, lunit = 2, ldtype = 4):
    '''
    Header structure, 
    nlines: header lines
    lname: variable name line
    lunit: variable unit line
    ldtype: variable datatype line
    '''
    df = df.copy()

    TYPE_MAP = {
        "IEEE4": np.float32,
        "IEEE8": np.float64,
        "LONG": np.int32,
        "ULONG": np.uint32,
        "FP2": np.dtype('>H'),  # big-endian unsigned short
    }

    with open(p_header, "rb") as f:
        headers = []
        for l in range(nlines):
            headers.append(f.readline())
        names = headers[lname].decode().strip().replace('"', "").split(",")
        if lunit:
            units = headers[lunit].decode().strip().replace('"', "").split(",")
        types = headers[ldtype].decode().strip().replace('"', "").split(",")
        dtype = np.dtype([(n, TYPE_MAP[t]) for n, t in zip(names, types)])
    
    big2little_names = pd.Index(deepcopy(names))
    if "FP2" in types:
        fp2_df = pd.DataFrame(zip(names, types), columns = ["name", "type"])
        fp2_names = fp2_df.loc[fp2_df["type"] == "FP2", "name"].tolist()
        # --------------------------------------------------------------------------------
        fp2vals = df[fp2_names].values
        if len(fp2vals[np.where(fp2vals > 9999.)]) > 0: raise Exception('Too large value!')
        # ----------------------------------------------------------------------------
        df.loc[:, fp2_names] = df.loc[:, fp2_names].applymap(val2FP2).applymap(FP22big)
        big2little_names = big2little_names.difference(fp2_names)
    df.loc[:, big2little_names] = df.loc[:, big2little_names].applymap(int2little)

    # ========================================================================================
    # write to file
    hex_headers = b"".join(headers)

    hex_body = []
    for cnt, row in df.iterrows():
        row = b"".join(row.to_list())
        hex_body.append(row)
    hex_body = b"".join(hex_body)
    hex_text = hex_headers + hex_body

    with open(p_save, "wb") as f:
        f.write(hex_text)



import numpy as np
import pandas as pd
from copy import deepcopy
from functools import reduce
from .specutils import numlist2arr

def SelectStructure(data, bins):
    data = deepcopy(data)
    binidx = np.where(bins != 99)[0]
    if isinstance(data, pd.DataFrame):
        return data.iloc[binidx, :]
    elif isinstance(data, dict):
        data_out = {}
        for key, val in data.items():
            if isinstance(val, pd.Series):
                data_out[key] = val[binidx]
            elif isinstance(val, pd.DataFrame):
                data_out[key] = val.iloc[binidx, :]
            elif isinstance(val, np.ndarray):
                data_out[key] = val[binidx, :]
            else:
                raise(Exception('Wrong type!'))
        return data_out
    else:
        raise(Exception('None correct type of data!'))

def CalcLambda(Gas, LPM, IR, P):
    # --------- Lambda value for tube attenuation --------------------
    # P     = tube pressure
    # LPM   = flow rate       (low limit of 0.1 lpm)
    # IR    = tube internal radius  (low limit of 0.000005 m)
    # Gas   = 'CO2', 'CH4', 

    LPM = numlist2arr(LPM)
    IR = numlist2arr(IR)
    P = numlist2arr(P)
    
    U  = LPM / (1000 * 60 * np.pi * IR**2)
    v  = (3.719815 - (0.0358737 * P) + (0.0001342 * P**2)) / 100000
    Re = (2 * IR * U) / v

    Dif = np.ones(U.shape)
    D0 = np.ones(U.shape)
    Da = np.ones(U.shape)
    Db = np.ones(U.shape)
    Dc = np.ones(U.shape)
    Dd = np.ones(U.shape)
    if Gas == 'CO2':
        Dif = Dif * 0.0000139
        D0  = D0 * 0.977616284023198
        Da  = Da * -40511.5062176583
        Db  = Db * 742694287.9160520
        Dc  = Dc * -3394928055533.75
        Dd  = Dd * 4991471292069380.0
    elif Gas == 'H2O':
        Dif = Dif * 0.0000239     
        D0  = D0 * 0.490851295350703
        Da  = Da * -12093.7897865916
        Db  = Db * 325341636.671577
        Dc  = Dc * -1502053321310.87
        Dd  = Dd * 2244127770065660.0
    elif Gas == 'CH4':
        Dif = Dif * 0.0000239
        D0  = D0 * 0.523609894456971
        Da  = Da * -13944.2741193251
        Db  = Db * 352263929.4320020
        Dc  = Dc * -1621482983243.50
        Dd  = Dd * 2416389615565820.0
    else:
        Dif = Dif * 0.0000239
        D0  = D0 * 0.977616284023198
        Da  = Da * -40511.5062176583
        Db  = Db * 742694287.9160520
        Dc  = Dc * -3394928055533.75
        Dd  = Dd * 4991471292069380.0

    idL = np.where(Re <= 2300)
    idT = np.where(Re >  2300)


    Lam = np.ones(U.shape)

    # calc non-turbulent lambda
    Lam[idL] = (0.0104 * v[idL] * Re[idL]) / Dif[idL]

    # calc turbulent lambda
    Lam[idT] = D0[idT] + Da[idT] / Re[idT] + Db[idT] / Re[idT]**2 + Dc[idT] / Re[idT]**3 + Dd[idT] / Re[idT]**4

    return [Lam, U, Re]

def TubeAtten(Gas,P,LPM,L,ID):
    # --------- Frequency response correction --------------------
    #  P     = tube pressure
    #  LPM   = flow rate       (low limit of 0.1 lpm)
    #  L     = tube length, m  (low limit of 0.1 m)
    #  ID    = tube internal diameter  (low limit of 0.00001 m)
    #  Gas   = 'CO2', 'CH4', 
    LPM = np.array(LPM)
    L = np.array(L)
    ID = np.array(ID)
    
    d4pi2  = (2 * 3.14159265)**2
    LPM[LPM <= 0.0] = 0.1
    L[L <= 0.0] = 0.1
    ID[ID <= 0.0] = 0.00001
    IR = ID / 2
    [Lam, U, Re] = CalcLambda(Gas, LPM, IR, P)
    Tatten = (d4pi2 * Lam * L * IR) / (U * U)
    return Tatten

def FreqResp2(Modelo, F, Zo, Do, Uo, ZLo, RLo, SSo, Pv1o, Pv2o, Ps1o, Ps2o, Tau1o, Tau2o, G1o, G2o):
    # --------- Frequency response correction --------------------
    #  Modelo = spectral type can be 'U', 'V', 'W', 'X', 'UW','UX' or 'WX'
    #  F     = frequency                (NxM vector)
    #  Zo    = height                   (Nx1 vector)
    #  Do    = zero plane displacement  (Nx1 vector)
    #  Uo    = wind speed               (Nx1 vector)
    #  ZLo   = stability                (Nx1 vector)
    #  RLo   = run length               (Nx1 vector)
    #  Pvo   = vector path length       (Nx1 vector)
    #  Pso   = scalar path length       (Nx1 vector)
    #  Tau1o = time constant            (Nx1 vector)
    #  Tau2o = time constant            (Nx1 vector)
    #  Go    = tube atten coef          (Nx1 vector)
    #  SSo   = sensor separation        (Nx1 vector)
    
    def repeat2vector(val, c1):
        val = numlist2arr(val).reshape(-1, 1)
        return np.tile(val, (1, c1))

    if F.ndim == 1: F = F.reshape(1, -1)
    [r1, c1] = F.shape
    
    Z = repeat2vector(Zo, c1)
    D = repeat2vector(Do, c1)
    U = repeat2vector(Uo, c1)
    ZL = repeat2vector(ZLo, c1)
    RL = repeat2vector(RLo, c1)
    Pv1 = repeat2vector(Pv1o, c1)
    Pv2 = repeat2vector(Pv2o, c1)
    Ps1 = repeat2vector(Ps1o, c1)
    Ps2 = repeat2vector(Ps2o, c1)
    Tau1 = repeat2vector(Tau1o, c1)
    Tau2 = repeat2vector(Tau2o, c1)
    G1 = repeat2vector(G1o, c1)
    G2 = repeat2vector(G2o, c1)
    SS = repeat2vector(SSo, c1)
    
    # calculate spectra
    Zi = 1000 * np.ones(Z.shape)
    Spec =  VectorModelSpectra(Modelo,F,ZL,U,Z,D,Zi)

    # calculate frequency response curve
    # BlockAvg
    A  = np.pi * F * RL * 60
    FR = 1 - (np.sin(A) / A)**2
    
    # Scalar Path 1
    if np.min(Ps1) > 0:
        A  = 2.0 * np.pi * F * Ps1 / U
        B  = np.exp(-1.0 * A)
        C  = (3 + B - (4 / A) * (1-B)) / A
        FR = FR * np.sqrt(C)
    # Scalar Path2
    if np.min(Ps2) > 0:
        A  = 2.0 * np.pi * F * Ps2 / U
        B  = np.exp(-1.0 * A)
        C  = (3 + B - (4 / A) * (1 - B)) / A
        FR = FR * np.sqrt(C)

    # Vector Path 1
    if np.min(Pv1) > 0:
        A = 2 * np.pi * F * Pv1 / U
        B = np.exp(-1 * A)
        C = (2 / A) * (2 + B - (3 * (1 - B) / A))
        FR = FR * np.sqrt(C)

    # Vector Path 2
    if np.min(Pv2) > 0:
        A = 2 * np.pi * F * Pv2 / U
        B = np.exp(-1 * A)
        C = (2 / A) * (2 + B - (3 * (1 - B) / A))
        FR = FR * np.sqrt(C)

    # Time Response 1
    if np.min(Tau1) > 0:
        A  = (2 * np.pi * F * Tau1)**2
        FR = FR * (1 / np.sqrt(1 + A))
 
    # Time Response 2
    if np.min(Tau2) > 0:
        A  = (2 * np.pi * F * Tau2)**2
        FR = FR * (1 / np.sqrt(1 + A))

    # Time Response differences
    if (np.min(Tau1) > 0) & (np.min(Tau2) > 0):
        A  = (2 * np.pi * F)**2
        FR = FR * ((1 + A * Tau1 * Tau2) / (np.sqrt(1 + A * Tau1**2) * np.sqrt(1 + A * Tau2**2)))

    # Tube Attenuation 1
    if np.min(G1) > 0:
        A = (F)**2
        B = np.exp(-1 * G1 * A)
        # If (B < 0.0000000001) Then B := 0.0000000001;
        FR = FR * B

    # Tube Attenuation 2
    if np.min(G2) > 0:
        A = (F)**2
        B = np.exp(-1 * G2 * A)
        # If (B < 0.0000000001) Then B := 0.0000000001;
        FR = FR * B

    # Sensor Separation
    if np.min(SS) > 0:
        A  = F * SS / U
        #  If (dA > 4.0) Then  dA := 4.0;
        if (A <= 4).all(): # ASK TIM, all or any?
            B = np.exp(-9.9 * A**1.5)
            FR = FR * B
    #
    SpecAtten = Spec * FR
    FRcor =  np.trapz(Spec, np.log(F)) / np.trapz(SpecAtten, np.log(F))

    return [FRcor, F, Spec, SpecAtten, FR]

def VectorModelSpectra(model,freq,ZOL,U,Z,Dz,Zi):
    # ------------------------------------------------------------------------
    #   S =  ModelSpectra(type,freq,ZOL,U,Z,Zi);

    #   ModelSpectra returns an array holding Kaimal model spectra or cospectra 
    #     passed parameters are:
    #         model  -  can be 'U', 'V', 'W', 'X', 'UW','UX' or 'WX'
    #         freq   - a vector of natural frequency values for which spectra
    #                  will be calculated
    #         ZOL    - Monin-Obhukov stability
    #           U    - Wind speed  in m/s
    #           Z    - Sensor height in m 
    #           Dz   - Canopy zero plane displacement in m
    #           Zi   - Boundary layer height in m

    # ------------------------------------------------------------------------
    
    # if (freq.ndim == 2) & (np.min(freq.shape) == 1):
    #     freq = freq.flatten()
    # get number of freq points and set size of model output
    # Notice here, all variavels are row vectors:
    freq = freq.ravel()
    ZOL = ZOL.ravel()
    U = U.ravel()
    Z = Z.ravel()
    Dz = Dz.ravel()
    Zi = Zi.ravel()
    
    S = np.zeros(freq.shape)

    # Calc normalized freq - handle no z/L or no Ubar
    #limit z/L values
    U[np.where(U <= 0.0)] = 0.01; 
    #limit z/L values
    ZOL[np.where(ZOL > 3.0)] = 3; 
    ZOL[np.where(ZOL < -3.0)] = -3; 

    freq = freq * ((Z-Dz) / U)

    # Find stable and unstable conditions
    idu = np.where(ZOL <= 0)
    ids = np.where(ZOL >  0)
    # calculate unstable U spectra   (Hojstrop 1981)
    if model == 'U': 
        # unstable U  
        A = -1 * ZOL[idu]**(2/3)
        B = (Z[idu] / Zi[idu])**(5/3)
        C = 9.546 + (1.235 * A / B**(2/5))
        C = 1 / C
        D = 210 * freq[idu]
        E = 1 + 33 * freq[idu]**(5/3)
        F = freq[idu] * A
        G = B + 2.2 * freq[idu]**(5/3)
        S[idu] = C * (D / E + F / G)
        # stable U
        A = 0.1676 + 0.2344 * ZOL[ids]
        B = 3.124 / (A**(2/3))
        C = A + B * freq[ids]**(5/3)
        S[ids] = freq[ids] / C

    # calculate unstable V spectra   (Hojstrop 1981)
    if model == 'V':
        # unstable V  
        aa = (Zi[idu] - Dz[idu]) / (Z[idu]-Dz[idu])   
        F = freq[idu] * aa
        A = (-1 * ZOL[idu] * aa)**(2/3)
        B = 0.32 * F * A;
        C = 1.0 + 1.1 * F**(5/3)
        D = 17 * freq[idu]
        E = (1 + 9.5 * freq[idu])**(5/3)
        S[idu] = B / C + D / E
        # stable V  
        A = ZOL[ids] * (0.74 + 4.7 * ZOL[ids]) / ((1 + 4.7 * ZOL[ids])**2)
        B = freq[ids] / (1.5 * A)
        C = 0.164 * B
        D = 1 + 0.164 * B**(5/3)
        S[ids] = C / D

    # calculate unstable W spectra  (Hojstrop 1981)
    if model == 'W':
        # unstable W  
        A = (-1 * ZOL[idu])**(2/3)
        B = 0.7285 + 1.4115 *A
        C = 1 / B
        D = freq[idu]
        E = 1 + 5.3 * freq[idu]**(5/3)
        F = 16 * freq[idu] * A
        G = (1 + 17 * freq[idu])**(5/3)
        S[idu] = C * (D / E + F / G)
        # stable W  
        A = 0.838 + 1.172 * ZOL[ids]
        B = 3.124 / (A**(2/3))
        C = A + B * freq[ids]**(5/3)
        S[ids] = freq[ids] / C

    # calculate  X spectra  (Moore 1981)
    if model == 'X': 
        # Find stable and unstable conditions
        iduLo = np.union1d(np.where(ZOL <= 0), np.where(freq <= 0.15))
        iduHi = np.union1d(np.where(ZOL <= 0), np.where(freq >  0.15))
        #  unstable X 
        A = 14.94 * freq[iduLo]
        B = (1 + 24 * freq[iduLo])**(5/3)
        S[iduLo] = A / B
        A = 6.827 * freq[iduHi]
        B = (1 + 12.5 * freq[iduHi])**(5/3)
        S[iduHi] = A / B
        #  unstable X 
        A = 0.0961 + 0.644 * ZOL[ids]**0.6
        B = 3.124 / (A**(2/3))
        C = A + B * freq[ids]**(5/3)
        S[ids] = freq[ids] / C

    # calculate UW spectra        - Moore 86 (Kaimal )
    if model == 'UW': 
        # Find stable and unstable conditions
        iduLo = np.union1d(np.where(ZOL <= 0), np.where(freq <= 0.24))
        iduHi = np.union1d(np.where(ZOL <= 0), np.where(freq >  0.24))
        #  unstable UW
        A = 20.78 * freq[iduLo]
        B = (1 + 31 * freq[iduLo])**1.575
        S[iduLo] = A / B
        A = 12.66 * freq[iduHi]
        B = (1 + 9.6 * freq[iduHi])**2.4;
        S[iduHi] = A / B
        # unstable UW
        A = 0.124 * (1 + 7.9 * ZOL[ids])**0.75
        B = 2.34 * A**(-1.1)
        C = A + B * freq[ids]**2.1
        S[ids] = freq[ids] / C

    # calculate unstable UX spectra
    if model == 'UX': 
        #  unstable UX
        A = 40 * freq[idu]
        B = (1 + 14 * freq[idu])**2.6
        S[idu] = A / B

        # calculate stable UX cospectra  Kaimal 73
        # cospec is normalized multiply by cov to compare with other cospec???????????????????
        A = ZOL[ids] * (0.74 + 4.7 * ZOL[ids]) / ((1 + 4.7 * ZOL[ids])**2)
        B = 2 * A
        C = (0.85 * freq[ids]) / (B + (1.7 * (freq[ids]**2.2) / (B**1.2)))
        S[ids] =  C

    # calculate unstable WX spectra      // Moore 86 (Kaimal 72)
    if model == 'WX': 
        # Find stable and unstable conditions
        iduLo = np.union1d(np.where(ZOL <= 0), np.where(freq <= 0.54))
        iduHi = np.union1d(np.where(ZOL <= 0), np.where(freq >  0.54))
        A = 12.92 * freq[iduLo]
        B = (1 + 26.7 * freq[iduLo])**1.375
        S[iduLo] = A / B
        A = 4.378 * freq[iduHi]
        B = (1 + 3.8 * freq[iduHi])**2.4
        S[iduHi] = A / B

        # calculate stable WX cospectra    Moore 86, after Kaimal 72
        A = 0.284 * (1 + 6.4 * ZOL[ids])**0.75
        B = 2.34 * A**(-1.1)
        C = A + B * freq[ids]**2.1
        S[ids] = freq[ids] / C

    return S.reshape(1, -1) # convert to row vector

def AvgSpectra(Grp,Frq,Spec,Cvar,Norm,fmin,fmax,bin_num,Slim,smean):
    # Remove nan data
    d = [
        np.where(~np.isnan(Grp))[0], 
        np.where(~np.isnan(Spec))[0],
        np.where(~np.isnan(Frq))[0],
        np.where(~np.isnan(Cvar))[0],
        np.where(np.abs(Cvar) >= Slim)[0]
    ]

    noNan = reduce(np.intersect1d, d)
    Gx = Grp[noNan]
    Frq = Frq[noNan,:]
    Spec = Spec[noNan,:]
    Cvar = Cvar[noNan]


    gkeys = np.sort(np.unique(Gx))

    # gkeys = sort(unique(Gx(:,1)));

    rg = gkeys.shape

    [rf, cf] = Frq.shape
    [rs, cs] = Spec.shape

    assert (rs == rf) & (cs == cf), 'mismatch in frequency and spectra matrix sizes'

    # make matrix of grouping values
    G1 = np.tile(Gx.reshape(-1, 1), (1, cf))

    # Make frequency key ranges
    logfmin = np.log10(fmin)
    logfmax = np.log10(fmax)

    # dlogn = (logfmax - logfmin)/bin_num
    # fkeys = np.arange(logfmin, logfmax, step = dlogn)
    fkeys = np.linspace(logfmin, logfmax, num = bin_num + 1, endpoint = True)

    fkey_edges = [-np.inf] + fkeys.tolist() + [np.inf]

    # take log10 of frequencies for averaging
    Frq = np.log10(Frq)

    if Norm == 'pre':
        Spec = Spec / np.tile(Cvar, (1, cs))


    freq_means = []
    freq_stds = []

    spec_means = []
    spec_stds = []

    counts = []

    idxs = []
    for i in range(len(fkeys) + 1):
        low = fkey_edges[i]
        up = fkey_edges[i + 1]
        idx = np.where((Frq > low) & (Frq < up))
        freq_means.append(np.nanmedian(Frq[idx]))
        freq_stds.append(np.nanstd(Frq[idx]))

        spec_means.append(np.nanmedian(Spec[idx]))
        spec_stds.append(np.nanstd(Spec[idx]))

        counts.append(len(idx[0]))
        idxs.append(low)

    #     break
    freq_means = np.array(freq_means).reshape(-1, 1)
    freq_stds = np.array(freq_stds).reshape(-1, 1)
    spec_means = np.array(spec_means).reshape(-1, 1)
    spec_stds = np.array(spec_stds).reshape(-1, 1)
    counts = np.array(counts).reshape(-1, 1)

    freq_means = 10**freq_means
    freq_stds = 10**freq_stds
    
    M = pd.DataFrame(
        np.hstack([freq_means, freq_stds, spec_means, spec_stds, counts]), 
        columns = ['Freq_mean', 'Freq_sd', 'Spec_mean', 'Spec_sd', 'Count'],
        index = idxs
    )
    
    return M


import numpy as np
import pandas as pd
from datetime import datetime, timedelta

def center_rolling(ts, rolling_window, method = 'mean'):
    # rolling_window = 5
    ts = ts.copy()
    assert rolling_window%2 == 1, "Not an odd rolling windown"
    shift_step = -int((rolling_window - 1) / 2)
    # ts = ts.rolling(rolling_window, min_periods = 1).median().shift(shift_step).ffill()
    if method == 'median':
        ts = ts.rolling(rolling_window).median().shift(shift_step)
    elif method == 'min':
        ts = ts.rolling(rolling_window).min().shift(shift_step)
    else:
        ts = ts.rolling(rolling_window).mean().shift(shift_step)
    return ts


def retuning(ts, window_size):
    # returning issue detection function v2
    ts = ts.copy()
    # timeseries ratio
    tsr = ((ts / center_rolling(ts, window_size)).abs() - 1) * 100 # 101: window = 10 seconds
    tsr = tsr.abs().resample("60S").min()
    tsr = tsr.resample("100ms").interpolate()
    
    anomaly = ts[tsr > 0.2]
    if not anomaly.empty:
        ano = anomaly.to_frame().reset_index()
        
        start = ano['datetime'].diff(1).map(
            lambda x: np.abs(x.total_seconds())
        )
        end = ano['datetime'].diff(-1).map(
            lambda x: np.abs(x.total_seconds())
        )
        
        tol = 30 # seconds
        end = end[end > tol].index.tolist()
        start = start[start > tol].index.tolist()
        start = [0] + start
        end = end + [len(ano) - 1]
        
        for s, e in zip(start, end):
            dif = (ano['datetime'][e] - ano['datetime'][s]).total_seconds() / 60
            if dif < 2: continue # minute
            s_dt = ano['datetime'][s]
            e_dt = ano['datetime'][e]
            # ts[s_dt: e_dt] = np.nan
            s_dt = s_dt - timedelta(minutes = 5)
            e_dt = e_dt + timedelta(minutes = 5)
            # ts[iqr_qc(ts[s_dt: e_dt], *stats4qc(ts[ts.index.difference(ts[s_dt:e_dt].index)]))] = np.nan
            ts[s_dt: e_dt] = np.nan
    
    """
    ts[tsr > 0.1] = np.nan
    
    if not ts[tsr > 0.1].empty:
    
        s = ts.index.get_loc(ts[tsr > 0.1].index[0])
        e = ts.index.get_loc(ts[tsr > 0.1].index[-1])

        ts[iqr_qc(ts, *stats4qc(ts[ts.index.difference(ts[s:e].index)]))] = np.nan
    """
    return ts

def qc_CO2(ts):    
    # ==================================================================================
    # retuning
    tsd = ts.copy()
    tsd_retuning = retuning(tsd, 601) # seconds
    # # local (hourly) detection:
    # [low, high] = stats4qc(tsd_retuning)
    # print(low, high)
    # tsd_retuning[iqr_qc(tsd_retuning, low, high)] = np.nan
    # print(tsd_retuning)
    # # global (daily) detection:
    # ==================================================================================
    # despiking
    # tsd_retuning = despike(tsd_retuning, threshold = 2)
    # tsd_retuning[despike(detrend(tsd_retuning)).isna()] = np.nan
    # # -----------------------------------------------------------------------------
    # # plotting
    # fig, axes = plt.subplots(3, 1, figsize = (10, 6))
    # axes = axes.flatten()
    # axes[0].scatter(tsd.index, tsd)
    # axes[1].scatter(tsd_retuning.index, tsd_retuning)
    # axes[0].sharex(axes[1])
    # axes[0].sharey(axes[1])
    # ax = axes[2]
    # ax.hist(tsd_retuning, bins = 300)
    # ax.axvline(low, c = 'red')
    # ax.axvline(high, c = 'red')
    # # -----------------------------------------------------------------------------
    # break
    # ==================================================================================
    # iqr control
    #     ts[iqr_qc(ts, *stats4qc(ts))] = np.nan
    return ts, tsd_retuning


# -*- coding: utf-8 -*-
"""
Created on Wed Dec 30 15:42:37 2020

@author: sz394@exeter.ac.uk
"""

import numpy as np
import pandas as pd
from pathlib import Path

class Raw():
    def __init__(self):
        # Variable type map, FP2 is the difficulty involving hex and endian conversion.
        self.TYPE_MAP = {
            "IEEE4": np.float32,
            "IEEE8": np.float64,
            "LONG": np.int32,
            "ULONG": np.uint32,
            "FP2": np.dtype('>H'),  # big-endian unsigned short
        }
        
    def __covert_fp2(self, val):
        """
        Covert big endian input val to normal value.

        Parameters
        ----------
        val : hex
            val is read as big endian.

        Returns
        -------
        val : float
            Normal float value.

        """
        pinf = 0x1fff
        ninf = 0x9fff
        anan = 0x9ffe
        # swap the bits (assume for small endian? sz at 2020-11-06)
        # val = ((val & 0xff00)>>8) | ((val & 0x00ff)<<8)
        # print(val)
        if ((val == pinf) or (val == ninf) or (val == anan)):
            val = np.nan
        else:
            # extract float components
            isNeg  =  not((val & 0x8000) == 0)
            Mantis =  val & 0x1FFF
            Expnt  =  (val & 0x6000) >> 13

            NV = Mantis
            NV = NV/(10.0**Expnt)

            if isNeg:
                NV = -1* NV
            val = NV
        return val

    def read(self, p, nlines = 5, lname = 1, lunit = 2, ldtype = 4):
        """
        Read Loggernet converted tob1 raw files.
        This version deals with FP2 data type.
        Parameters
        ----------
        p : str or windows path
            The location of tob1 raw file.
        nlines : int, optional
            Lines number of meta info. The default is 5.
        lname : int, optional
            Line number of variable names. The default is 1.
        lunit : int, optional
            Line number of variable units, can be 0. The default is 2.
        ldtype : int, optional
            Line number of variable dtype. The default is 4.

        Returns
        -------
        df : TYPE
            DESCRIPTION.

        """
        with open(p, "rb") as f:
            headers = []
            for l in range(nlines):
                headers.append(f.readline())
            names = headers[lname].decode().strip().replace('"', "").split(",")
            if lunit:
                units = headers[lunit].decode().strip().replace('"', "").split(",")
            types = headers[ldtype].decode().strip().replace('"', "").split(",")
            dtype = np.dtype([(n, self.TYPE_MAP[t]) for n, t in zip(names, types)])

            array = np.fromfile(f, dtype = dtype, count = -1)
            df = pd.DataFrame(array)

        if "FP2" in types:
            fp2_df = pd.DataFrame(zip(names, types), columns = ["name", "type"])
            fp2_names = fp2_df.loc[fp2_df["type"] == "FP2", "name"].tolist()
            df.loc[:, fp2_names] = df.loc[:, fp2_names].applymap(self.__covert_fp2)
            
        if "SECONDS" in  df.columns:
            fdel = df["SECONDS"].astype(np.int64) * 1000000000
            if "NANOSECONDS" in df.columns:
                fdel += df["NANOSECONDS"].astype(np.int64)

            datetime = np.datetime64('1990-01-01') + fdel.astype("timedelta64[ns]")
            datetime.name = "datetime"
            df.index = datetime
        if lunit:
            df.attrs["unit"] = dict(zip(names, units))
            # Alternative way that might trigger warnings
            # df.unit = dict(zip(names, units))
        return df

    def extract(self, df, col_names, rename_cols = None, nanval = None):
        dfo = df[col_names]
        if nanval:
            dfo = dfo.replace(nanval, np.nan)
        if rename_cols:
            assert type(rename_cols) == dict, "rename_cols type must be dict" 
            dfo = dfo.rename(columns = rename_cols)
        return dfo


import numpy as np
import pandas as pd
from pathlib import Path

def unit2unit(val, choice):
    """
    Input:
    ------
    val:
        value of which the unit to be changed. 
    choice: int
        what unit change to what unit?
        g/kg -> kg/kg
    """
    if (choice == "g/kg -> kg/kg") or (choice == 1):
        val = val / 1000
    elif (choice == "g/m3 -> mg/m3") or (choice == 2):
        val = val * 1000
    else:
        raise(Exception("Must specify a correct choice"))
    return val

def fetch_sort_path(root_proj = None, inputlist = None, return_time = True):
    if root_proj:
        paths = root_proj.glob("*.dat")
        paths = list(paths)
    elif inputlist:   
        with open(inputlist, 'r') as f:
            paths = f.readlines()
        paths = [Path(p.strip()) for p in paths]
    else:
        raise Exception('No input list or data directory!')
    # --------------------------------------------------------------------------------------------------
    path_time = [pd.to_datetime("".join(p.stem.split("_")[-4::]), format = "%Y%m%d%H%M") for p in paths]
    paths = [p for _, p in sorted(zip(path_time, paths))]
    path_time = sorted(path_time)
    # print(path_time)
    if return_time:
        return paths, path_time
    else:
        return paths
    
def group_day(paths, path_time):
    groups = pd.DataFrame(
        zip([p.year for p in path_time], [p.month for p in path_time], [p.day for p in path_time], paths), 
        columns = ['year', 'month', 'day', 'path']
    ).groupby(['year', 'month', 'day'])
    
    return groups


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from pathlib import Path
from datetime import datetime, timedelta
from .qcutils import stats4qc, iqr_qc, glbdec, despike, detrend, deplatform
from .CO2 import center_rolling, retuning, qc_CO2
from .Ts import get_minute_std, qc_Ts
from .mutual import two_systems

__all__ = [
    'np', 'pd', 'plt', 'Path', 'datetime', 'timedelta',
    'stats4qc', 'iqr_qc', 'glbdec', 'despike', 'detrend', 'deplatform',
    'center_rolling', 'retuning', 'qc_CO2',
    'get_minute_std', 'qc_Ts',
    'two_systems'
    ]


import numpy as np
from collections import namedtuple
from scipy import signal, integrate


CorrectedWind3D = namedtuple("CorrectedWind3D",
                             ["uvw_ship",
                              "euler_angles",
                              "euler_angles_angular_rates",
                              "euler_angles_accelerations",
                              "euler_angles_slow",
                              "euler_angles_fast",
                              "mount_offset_rotations",
                              "uvw_earth",
                              "uvw_angular",
                              "uvw_linear",
                              "ship_enu",
                              "uvw_enu",
                              "imu_enu"])

def rotation_matrix(theta, axis, active=False):
    """Generate rotation matrix for a given axis
    The default (``active=False``) output matrix corresponds to passive
    ("alias") rotations, if used to post-multiply row vectors.
    Parameters
    ----------
    active: bool, optional
        Whether to return active transformation matrix.
    Note
    ----
    See ``rotate_coordinates`` for rest of parameters.
    Returns
    -------
    R_theta : numpy.ndarray
        3x3 rotation matrix
    """
    theta = np.radians(theta)
    if axis == 0:
        R_theta = np.array([[1, 0, 0],
                            [0, np.cos(theta), -np.sin(theta)],
                            [0, np.sin(theta), np.cos(theta)]])
    elif axis == 1:
        R_theta = np.array([[np.cos(theta), 0, np.sin(theta)],
                            [0, 1, 0],
                            [-np.sin(theta), 0, np.cos(theta)]])
    else:
        R_theta = np.array([[np.cos(theta), -np.sin(theta), 0],
                            [np.sin(theta), np.cos(theta), 0],
                            [0, 0, 1]])
    if active:
        R_theta = np.transpose(R_theta)
    return R_theta

def rotate_coordinates(vectors, theta, axis=2, **kwargs):
    """Rotate vector coordinate system or vectors themselves around an axis
    A right-handed coordinate system is assumed, where positive rotations
    are clockwise when looking along the rotation axis from the origin
    towards the positive direction.  With the default ``active=False``,
    each row vector :math:`i` in input coordinate system :math:`0` is
    rotated around the given axis by angle :math:`\\theta`, so that it can
    be expressed in coordinate system :math:`1` as follows:
    .. math:: \\vec{v}_{i,1} = \\vec{v}_{i,0} R_{\\theta}
    where the input and output coordinate systems are different.  If the
    input and output coordinate systems are the same, i.e. ``active=True``,
    the vectors are transformed according to:
    .. math:: \\vec{v}_{i,1} = \\vec{v}_{i,0} R_{\\theta}^\\intercal
    Parameters
    ----------
    vectors : array_like
        An nx3 array of vectors with their `x`, `y`, `z` components
    theta : numeric
        The angle (degrees) by which to perform the rotation.
    axis : int
        Axis around which to perform the rotation (`x` = 0; `y` = 1;
        `z` = 2).
    **kwargs
        Optional keywords
        active : bool
            Whether to return the coordinates of each vector in the rotated
            coordinate system or to rotate the vectors themselves onto the
            same coordinate system.  Default is to perform a passive
            transformation, i.e. rotation of the coordinate system.
    Returns
    -------
    array_like
        The vector array with the same shape as input with rotated
        components.
    """
    R_theta = rotation_matrix(theta, axis, **kwargs)
    # Multiplying a vector by R_theta, as defined above, transforms the
    # vector coordinates from the original coordinate system to the new,
    # rotated coordinate frame.  Note that the rotation matrix
    # post-multiplies the vector matrix.  If rotating the vectors
    # themselves, then the *transposed* rotation matrix post-multiplies the
    # matrix of vectors.
    return np.dot(vectors, R_theta)

def _rot_xyz(phi, theta, psi):
    """Compute passive rotation matrix from 3 angles in 1-D array x
    The matrix is meant to post-multiply row vectors.
    Parameters
    ----------
    phi : float
        Angle (degrees) for the rotation around the `x`-axis.
    theta : float
        Angle (degrees) for the rotation around the `y`-axis.
    psi : float
        Angle (degrees) for the rotation around the `z`-axis.
    Returns
    -------
    numpy.ndarray
        2-D array (3x3) with rotation matrix corresponding to the rotation
        sequence :math:`\\phi`, :math:`\\theta`, :math:`\\psi`.
    """
    rotx = rotation_matrix(phi, 0)
    roty = rotation_matrix(theta, 1)
    rotz = rotation_matrix(psi, 2)
    return np.transpose(np.dot(rotz, np.dot(roty, rotx)))


def euler_rotate(xyz, xyz_angles):
    """Rotate vectors using Euler angles
    Parameters
    ----------
    xyz : array_like
        2-D array (Nx3) with row vectors to be rotated.
    xyz_angles : array_like
        2-D array (Nx3) with row vectors of angles (degrees) to be used for
        rotation of ``xyz``, where the first column specifies the angle for
        the first rotation around the `x`-axis, the second column the angle
        for the second rotation around the `y`-axis, and the third column
        the angle for the last rotation around the `z`-axis.
    Returns
    -------
    numpy.ndarray
        Array of the same shape as ``xyz`` with rotated vectors.
    """
    xyz_rots = np.empty_like(xyz)
    for i, v in enumerate(xyz_angles):
        phi, theta, psi = v
        euler_mati = _rot_xyz(phi, theta, psi)
        xyz_roti = np.dot(xyz[i], euler_mati)
        xyz_rots[i] = xyz_roti
    return xyz_rots


def euler_rate_rotate(euler_angles, omega):
    """Rotate angular rates, given Euler angles
    The Euler angular rate column vector :math:`\\Omega` consists of the
    angular rates :math:`\\dot \\phi`, :math:`\\dot \\theta`, and
    :math:`\\dot \\psi` around the axes in the inertial frame.  It is
    related to the angular rate column vector :math:`\\omega`, consisting
    of the angular rates :math:`\\omega_x`, :math:`\\omega_y`, and
    :math:`\\omega_z` in the body frame via a non-orthogonal transformation
    matrix :math:`R`:
    .. math:: \\Omega = R \\omega
    This function performs this operation by using the transpose of the
    input body-frame angular rate row vectors, and transposing the result
    to return row vectors also.
    Parameters
    ----------
    euler_angles : array_like
        2-D array (Nx2) with Euler :math:`\\phi` (roll) and :math:`\\theta`
        (pitch) angles (radians) around the `x`- and `y`-axes in columns 1
        and 2, respectively, representing the orientation of the sensor
        orientation relative to the output frame.
    omega : float 2-D array
        (Nx3) with angular rates :math:`\\omega_x`, :math:`\\omega_y`, and
        :math:`\\omega_z` around the body-frame `x`-, `y`-, and `z`-axes in
        columns 1, 2, and 3, respectively.
    Returns
    -------
    array_like
        Array with same shape as `omega` with rotated angular rates.
    """
    def rot_mat(phi, theta):
        """Transformation matrix pre-multiplying column vectors"""
        return np.array([[1, np.sin(phi) * np.tan(theta),
                          np.cos(phi) * np.tan(theta)],
                         [0, np.cos(phi), -np.sin(phi)],
                         [0, np.sin(phi) / np.cos(theta),
                          np.cos(phi) / np.cos(theta)]])
    omega_rots = np.empty_like(omega)
    for i, v in enumerate(euler_angles):
        phi, theta = v
        # Multiply and transpose to original shape as row vectors
        omega_rots[i] = np.dot(rot_mat(phi, theta), omega[i].T).T
    return omega_rots


def _cumtrapz(x, sample_freq):
    """Cumulatively integrate signal given sampling frequency
    Parameters
    ----------
    x : array_like
        A 1- or 2-D array with signal such as angular rate (deg/s) or
        linear rate measurements.
    sample_freq : int, float
        The sampling frequency in units required for integration.
    Returns
    -------
    array_like
        Array with same shape as `x`
    """
    y = integrate.cumtrapz(x, dx=(1.0 / sample_freq), axis=0, initial=0)
    return y


def _butterworth_coefs(cutoff_period, sample_freq, Astop=10.0, Apass=0.5):
    """Compute Butterworth filter coefficients
    Parameters
    ----------
    cutoff_period : int, float
        Cutoff period (s) defining the cutoff frequency for designing the
        filter.
    sample_freq : int, float
        The sampling frequency in units required for integration.
    Apass : float, optional
        Stopband attenuation
    Astop : float, optional
        Passband ripple (dB)
    Returns
    -------
    tuple
        Tuple with (index, name in brackets):
        numpy.ndarray [0]
            Numerator (b) polynomial of the filter.
        numpy.ndarray [1]
            Denominator (a) polynomial of the filter.
        float [2]
            Padding length for the filter.
    """
    # Passband and stopband cutoffs, normalized to Nyquist frequency
    wp = 1.0 / (2.0 * cutoff_period) / (sample_freq / 2.0)
    ws = 1.0 / cutoff_period / (sample_freq / 2.0)
    N, Wn = signal.buttord(wp, ws, Apass, Astop)
    bc, ac = signal.butter(N, Wn, "high")
    # WATCH THIS: we need to make padlen the same as in Matlab
    pdl = 3 * (max(len(ac), len(bc)) - 1)
    return bc, ac, pdl

def wind3D_correct(wind_speed, acceleration, angle_rate, heading, speed,
                   anemometer_pos, sample_freq, Tcf, Ta,
                   tilt_motion=np.array([0.0, 0.0]),
                   tilt_anemometer=np.array([0.0, 0.0])):
    """Correct wind vector measurements from a moving platform
    This is a port of S. Miller's ``motion`` Matlab function, which
    implements Miller et al. (2008) approach.  Coordinate frame is assumed
    to be right-handed:
    * `x` - positive towards the bow.
    * `y` - positive towards port side.
    * `z` - positive upwards.
    Parameters
    ----------
    wind_speed : numpy.ndarray
        A 2-D (Nx3) array with `u`, `v`, and `w` wind speed (m/s) vectors.
    acceleration : numpy.ndarray
        A 2-D (Nx3) array with `x`, `y`, and `z` acceleration (m/s/s)
        vectors.
    angle_rate : numpy.ndarray
        A 2-D (Nx3) array with angular rates (radians/s) vectors.
    heading : array_like
        A vector with ship heading measurements (deg) in right-hand
        coordinate system.
    speed : array_like
        A vector with ship speed (m/s).
    anemometer_pos : array_like
        [`x`, `y`, `z`] position vector of anemometer, relative to motion
        sensor (m).
    sample_freq : float
        Sampling frequency (Hz).
    Tcf : float
        Complimentary filter period (s).
    Ta : float
        High-pass filter cutoff for accelerations (s).  This can be a
        scalar if the same cutoff is used for the three components, or a 3
        element vector to indicate cutoff period for the `x`, `y`, `z`
        components.
    tilt_motion : array_like, optional
        [roll, pitch] vector with mean tilt (radians) relative to a
        horizontal plane.  Roll angle offset is positive is port side up,
        and pitch is positive for bow down (see Miller 2008 for info, set
        to [0 0] if unknown).
    tilt_anemometer : array_like, optional
        [roll, pitch] vector with mean tilt (radians) relative to a
        horizontal plane.  Roll angle offset is positive is port side up,
        and pitch is positive for bow down (see Miller 2008 for info, set
        to [0 0] if unknown).
    Returns
    -------
    CorrectedWind3D : namedtuple
        namedtuple with (index, name in brackets):
        numpy.ndarray [0, `uvw_ship`]
            2-D array (Nx3) with corrected wind vectors in ship-referenced
            frame with z-axis parallel to gravity.
        numpy.ndarray [1, `euler_angles`]
            2-D array (Nx3) with Euler angles.
        numpy.ndarray [2, `euler_angles_angular_rates`]
            2-D array (NX3) with Euler angles from rate sensors
            (unfiltered).
        numpy.ndarray [3, `euler_angles_accelerations`]
            2-D array (NX3) with Euler angles from accelerometers
            (unfiltered).
        numpy.ndarray [4, `euler_angles_slow`]
            2-D array (NX3) with slow Euler angles (low pass filtered).
        numpy.ndarray [5, `euler_angles_fast`]
            2-D array (NX3) with fast Euler angles (high pass filtered).
        numpy.ndarray [6, `mount_offset_rotations`]
            2-D array (3X3) with mounting offset rotation matrix (see
            Miller 2008).
        numpy.ndarray [7, `uvw_earth`]
            2-D array (NX3) with measured velocity, rotated to the earth
            frame.
        numpy.ndarray [8, `uvw_angular`]
            2-D array (NX3) with velocity induced by angular motion.
        numpy.ndarray [9, `uvw_linear`]
            2-D array (NX3) with velocity induced by platform linear
            motion.
        numpy.ndarray [10, `ship_enu`]
            2-D array (NX3) with ship velocity in eastward, northward, up
            frame.  (low pass filtered).
        numpy.ndarray [11, `uvw_enu`]
            2-D array (NX3) with corrected wind in eastward, northward, up
            frame.
        numpy.ndarray [12, `imu_enu`]
            2-D (NX3) array with displacement of the motion sensor.
    References
    ----------
    Miller,S.D., Hristov,T.S., Edson,J.B., and C.A. Friehe, 2008:
        Platform Motion Effects on Measurements of Turbulence and Air-Sea
        Exchange Over the Open Ocean, J. Atmo. Ocean. Tech. 25(9),
        1683-1694, DOI: 10.1175/2008JTECHO547.1.
    """
    Astop, Apass = 10.0, 0.5     # stopband attenuation; passband Ripple (dB)
    # Stop, passband cutoffs
    bc, ac, pdl = _butterworth_coefs(Tcf, sample_freq,
                                     Astop=Astop, Apass=Apass)

    # EULER ANGLES: (see Edson et al., 1998)
    # Low frequency tilt using accelerometers and gyro
    g = np.sqrt(np.sum(np.mean(acceleration, 0) ** 2))
    gyro = np.unwrap(-np.radians(heading))  # Put heading in radians and RHS
    gyro0 = gyro[0]               # Initial heading
    gyro = gyro - gyro0           # Remove initial heading
    # Uncorrected Euler angles from accelerometers and gyro
    EA_acc = np.column_stack((np.arcsin(acceleration[:, 1] / g),  # phi
                              np.arcsin(-acceleration[:, 0] / g),  # theta
                              gyro))                               # psi
    # High frequency angles using angular rates
    rm = signal.detrend(angle_rate, 0, "constant")
    EA_rate = _cumtrapz(rm, sample_freq)  # Estimated Euler angles
    EA_rate_hf = signal.filtfilt(bc, ac, EA_rate, padlen=pdl, axis=0)
    theta_lf = (EA_acc[:, 1] -
                signal.filtfilt(bc, ac, EA_acc[:, 1], padlen=pdl))
    theta = theta_lf + EA_rate_hf[:, 1]
    phi_pitched = EA_acc[:, 0] / np.cos(theta_lf)
    phi_lf = phi_pitched - signal.filtfilt(bc, ac, phi_pitched, padlen=pdl)
    phi = phi_lf + EA_rate_hf[:, 0]  # complementary-filtered phi
    # Low-pass filter to retain low-frequency content, but not the offset
    psi_lf = gyro - signal.filtfilt(bc, ac, gyro, padlen=pdl)
    axg = np.arctan2(-acceleration[:, 0] * np.cos(phi), g)
    theta_lf = axg - signal.filtfilt(bc, ac, axg, padlen=pdl)
    theta = theta_lf + EA_rate_hf[:, 1]
    EA_slow = np.column_stack((phi_lf, theta_lf, psi_lf))

    # NONLINEAR ROTATION OF ANGULAR RATES FROM MOTION SENSOR-FRAME TO
    # EARTH-FRAME This next step puts the measured angular rates into the
    # earth reference frame.  Motion sensor frame-based angular rates are
    # first rolled based on the roll angle (phim) estimated above; they are
    # then pitched *in a rolled frame* based upon the pitch angle (thetam)
    # estimated above; finally, they are yawed *in a rolled and pitched
    # frame* based upon both the pitch and roll (thetam and phim)
    # estimates. These sequential rotations are exact for small angle RATES
    # (see Goldstein) - I've found this step doesn't have much effect on
    # the corrected winds.
    omega = euler_rate_rotate(np.column_stack((phi, theta)), rm)
    # The angular rates omega are now in the earth based frame, which is
    # what we want for creating the rotation-transformation
    # matrix. Re-integrate and high pass filter these angular rates to get
    # the fast angles.
    EA_fast = signal.filtfilt(bc, ac,
                              _cumtrapz(omega, sample_freq),
                              padlen=pdl, axis=0)

    # combine high- and low-frequency angles; make 2D transformation matrix
    EA = EA_slow + EA_fast

    # SHIP AND ANEMOMETER OFFSET ANGLES
    # Pitch and roll tilt of platform w.r.t. earth (see Wilczak et al. 2000)
    phi_em, theta_em = tilt_motion[0], tilt_motion[1]
    PHI_em = rotation_matrix(np.degrees(phi_em), axis=0, active=True)
    THETA_em = rotation_matrix(np.degrees(theta_em), axis=1, active=True)
    # Pitch and roll tilt of anemometer w.r.t. earth
    phi_ea, theta_ea = tilt_anemometer[0], tilt_anemometer[1]
    PHI_ea = rotation_matrix(np.degrees(phi_ea), axis=0, active=True)
    THETA_ea = rotation_matrix(np.degrees(theta_ea), axis=1, active=True)

    # Anemometer WRT motion sensor
    M_em = np.dot(THETA_em, PHI_em)
    M_ea = np.dot(THETA_ea, PHI_ea)
    M_ma = np.dot(np.transpose(M_em), M_ea)

    # WIND VECTOR ROTATION
    # Using coordinate rotations
    EA_degs = np.degrees(EA)    # only for getting rotation matrix
    Ur = euler_rotate(np.dot(M_ma, wind_speed.T).T, EA_degs)

    # PLATFORM ANGULAR VELOCITY
    uam = np.column_stack((anemometer_pos[2] * rm[:, 1] -
                           anemometer_pos[1] * rm[:, 2],
                           anemometer_pos[0] * rm[:, 2] -
                           anemometer_pos[2] * rm[:, 0],
                           anemometer_pos[1] * rm[:, 0] -
                           anemometer_pos[0] * rm[:, 1]))
    Ua = euler_rotate(uam, EA_degs)  # rotate to earth frame

    # PLATFORM LINEAR VELOCITY
    # High-pass filters for accelerometers
    if np.isscalar(Ta):
        Ta = np.repeat(Ta, 3)
    bax, aax, pdlx = _butterworth_coefs(Ta[0], sample_freq,
                                        Astop=Astop, Apass=Apass)
    bay, aay, pdly = _butterworth_coefs(Ta[1], sample_freq,
                                        Astop=Astop, Apass=Apass)
    baz, aaz, pdlz = _butterworth_coefs(Ta[2], sample_freq,
                                        Astop=Astop, Apass=Apass)
    # Rotate accelerations
    ae = euler_rotate(acceleration, EA_degs)
    ae[:, 2] = ae[:, 2] - g     # subtract gravity
    up = _cumtrapz(ae, sample_freq)
    Up = np.column_stack((signal.filtfilt(bax, aax, up[:, 0], padlen=pdlx),
                          signal.filtfilt(bay, aay, up[:, 1], padlen=pdly),
                          signal.filtfilt(baz, aaz, up[:, 2], padlen=pdlz)))

    # SHIP SPEED: this is in the positive x-direction (assuming the bow is
    # positive x direction).  We low-pass filter this signal so there isnt
    # double counting of linear motion from the integrated accelerometers.
    # Also, the y- and z-components are zero (ie, captured by the
    # integrated accelerometers).
    n = len(speed)
    Us = np.column_stack((speed -
                          signal.filtfilt(bax, aax, speed, padlen=pdlx),
                          np.zeros((n, 1)), np.zeros((n, 1))))

    UVW = Ur + Ua + Up + Us     # corrected wind vector

    # Organize outputs
    u_ea = np.column_stack((np.zeros((n, 1)),
                            np.zeros((n, 1)),
                            (gyro0 + np.pi / 2) * np.ones((n, 1))))
    u_ea_degs = np.degrees(u_ea)
    U_ship = euler_rotate(Us, u_ea_degs)
    U_earth = euler_rotate(UVW, u_ea_degs)
    # Platform displacement
    xp = _cumtrapz(Up, sample_freq)
    Xp = np.column_stack((signal.filtfilt(bax, aax, xp[:, 0], padlen=pdlx),
                          signal.filtfilt(bay, aay, xp[:, 1], padlen=pdly),
                          signal.filtfilt(baz, aaz, xp[:, 2], padlen=pdlz)))

    return CorrectedWind3D(UVW, EA, EA_rate, EA_acc, EA_slow, EA_fast,
                           M_ma, Ur, Ua, Up, U_ship, U_earth, Xp)


# -*- coding: utf-8 -*-
"""
Created on Thu Dec 31 14:45:24 2020

@author: sz394@exeter.ac.uk
"""

import numpy as np
# ==================================================================== #
# Resource links                                                       #
# Gas variable conversion:                                             #
# https://www.gastec.co.jp/en/technology/knowledge/concentration/      #
# Something about water vapor:                                         #
# https://www.cnblogs.com/ruo-li-suo-yi/p/7772777.html                 #
# Online converter to check algorithm:                                 #
# https://www.cactus2000.de/uk/unit/masshum.shtml                      #
# engineering toolbox                                                  #
# https://www.engineeringtoolbox.com/                                  #
# ==================================================================== #

# -------------------------------------------------------------------
# Constants
R = universal_gas_constant = 8.314 # J/K/mol
Md = dry_air_molecular_weight = 28.96546e-3 # unit: kg / mol
MH2O = Mw = water_molecular_weight = 18.015268 # unit: g / mol
MCO2 = carbon_dioxide_molecular_weight =  44.009 # unit: g/mol
NA = 6.0221409e23 # Avogadro number
# -------------------------------------------------------------------
# Basic equations
def percent2ppm(percent):
    """
    math: ppm=%×10000 → 1ppm=0.0001%
    """
    ppm = percent * 10000
    return ppm

def ppm2percent(ppm):
    """
    math: %=ppm×1/10000 → 1%=10000ppm
    """
    percent = ppm / 10000
    return percent
def ppm2mass_density(ppm, M, T, P = 1013.25):
    """
    ppm → mg/m3
    Input:
    ------
    ppm(parts per million):
        This unit that expresses concentration in parts per million is measured as the volume (denoted in litres [L]) of a substance found in 1L of a medium such as air.
    M(g/mol): 
        Molecular weight of a substance
    T(degC):
        Temperature
    P(hPa): 
        P denotes the atmospheric pressure at the point of measurement (hPa)
    Output:
    -------
    mass density (mg/m3)
        This unit expresses the concentration in one cubic metre of air (equivalent to 1L or 1000mL) of a substance in terms of its mass (measured in milligrams). It is primarily used for particle-like substances, and only rarely for gaseous concentrations.
    Note:
    -----
    22.4(L): 
        The volume of 1 mol at 1 atmospheric pressure at 0 degC
    273(K): 
        FK stands for Kelvin, the unit used to measure thermodynamic temperature; as 0 degC corresponds to 273.15K.
        You simply need to add 273 to the Celsius/Centigrade value (273+T) to obtain the temperature in Kelvin
    1013(hPa): 
        One atmospheric pressure    
    math: 
    -----
    mg/m3=ppm×(M/22.4)×(273/(273+T))×(P/1013)
    """
    mass_density = ppm * (M / 22.4)  * (273.15 / (273.15 + T)) * (P / 1013.25)
    return mass_density

def mass_density2ppm(mass_density, M, T, P = 1013.25):
    """
    mg/m3 → ppm
    Input:
    ------
    mass_density (mg/m3)
        This unit expresses the concentration in one cubic metre of air (equivalent to 1L or 1000mL) of a substance in terms of its mass (measured in milligrams). It is primarily used for particle-like substances, and only rarely for gaseous concentrations.
    M(g/mol): 
        Molecular weight of a substance
    T(degC):
        Temperature
    P(hPa): 
        P denotes the atmospheric pressure at the point of measurement (hPa)
    Output:
    -------
    ppm(parts per million):
        This unit that expresses concentration in parts per million is measured as the volume (denoted in litres [L]) of a substance found in 1L of a medium such as air.
    Note:
    -----
    22.4(L): 
        The volume of 1 mol at 1 atmospheric pressure at 0 degC
    273(K): 
        FK stands for Kelvin, the unit used to measure thermodynamic temperature; as 0 degC corresponds to 273.15K.
        You simply need to add 273 to the Celsius/Centigrade value (273+T) to obtain the temperature in Kelvin
    1013(hPa): 
        One atmospheric pressure    
    math: 
    -----
    ppm=mg/m3×(22.4/M)×((273+T)/273)×(1013/P)
    mg/m3=ppm×(M/22.4)×(273/(273+T))×(P/1013)
    """
    ppm = mass_density * (22.4 / M) * ((273.15 + T) / 273.15 ) * (1013.25 / P)
    return ppm

def volumn_density2percent(volumn_density, M, T, P = 1013.25):
    """
    mg/L → %
    Input:
    ------
    volumn_density(mg/L (milligram per litre)):
        This unit expresses the concentration in one litre of air (1000mL) of a substance in terms of its mass (measured in milligrams). It is generally used for measuring concentrations in liquids, and only rarely for gaseous concentrations.
    M(g/mol): 
        Molecular weight of a substance 
    T(degC):
        Temperature
    P(hPa): 
        P denotes the atmospheric pressure at the point of measurement (hPa)
    Output:
    -------
    % (percent):
        This unit expresses concentration in parts per hundred (percentage) of a substance in 100mL of a medium such as air.
    math:
    -----
    %=mg/L×(22.4/M)×((273+T)/273)×(1/10)×(1013/P)
    """
    percent = volumn_density * (22.4 / M) * ((273.15 + T) / 273.15 ) * (1 / 10) * (1013.25 / P)
    return percent

def percent2volumn_density(percent, M, T, P = 1013.25):
    """
    % → mg/L
    Input:
    ------
    % (percent):
        This unit expresses concentration in parts per hundred (percentage) of a substance in 100mL of a medium such as air.
    M(g/mol): 
        Molecular weight of a substance 
    T(degC):
        Temperature
    P(hPa): 
        P denotes the atmospheric pressure at the point of measurement (hPa)
    Output:
    -------
    volumn_density(mg/L (milligram per litre)):
        This unit expresses the concentration in one litre of air (1000mL) of a substance in terms of its mass (measured in milligrams). It is generally used for measuring concentrations in liquids, and only rarely for gaseous concentrations.
    math:
    -----
    mg/L=%×(M/22.4)×(273/(273+T))×10×(P/1013)
    """
    volumn_density = percent * (M/ 22.4) * (273.15 / (273.15 + T)) * 10 * (P / 1013.25)
    return volumn_density

def saturation_vapor_pressure(T, es0 = 6.112):
    """
    Input:
    ------
    T(degC):
        Temperature
    es0 (hPa):
        saturation_vapor_pressure at 0 degC
    Output:
    -------
    saturation_vapor_pressure(hPa)
    math:
    -----
    Tetens formula
    6.112 e^((17.67 * T) / (T + 243.5)
    """
    es = 6.112 * np.e ** ((17.67 * T) / (T + 243.5))
    return es

def relative_humidity2dewpoint(T, RH):
    """
    Input:
    ------
    T(degC):
        Temperature
    RH:
        Relative humidity in the range [0 - 1]
    Output:
    -------
    Td(degC):
        Dewpoint temperature
    math:
    -----
    Magnus-Tetens Approximation
    Valid when:
        0 < T < 60
        1% < RH < 100%
        0 degC < Td < 50 degC
    """
    a = 17.27 # constant, unitless
    b = 237.7 # constant, degC
    gamma = (a * T) / (b + T) + np.log(RH) # np.log is ln
    Td = (b * gamma) / (a - gamma)
    return Td

def vapor_pressure2absolute_humidity(e, T):
    """
    Input:
    ------
    e(hPa): 
        vapor pressure
    T(degC):
        temperature
    Output:
    -------
    absolute_humidity(g/m3)
    Note:
    -----
    Rv: water vapor gas constant 461.5(J Kg^-1 K^-1)
    """
    Rv = 461.5
    # convert T from degC to Kelvin
    # convert e from hPa to Pa
    absolute_humidity = (e * 100)/(Rv * (T + 273.15)) # Unit: Kg/m3
    # Kg/m3 -> g/m3
    absolute_humidity = 1000 * absolute_humidity
    return absolute_humidity 

 
def specific_humidity(e, P):
    """
    Input:
    ------
    e(hPa): 
        vapor pressure
    P(hPa):
        atmospheric pressure
    Output:
    -------
    q(g/kg):
        Specific humidity
    math:
    -----
    q = mv / (mv + md) = mv/m = rhov / rho = (622 * e) / (p - 0.378 * e) g/Kg = (0.622 * e) / (P - 0.378 * e) g/g
    where
        mv: mass of vapor
        md: mass of dry air
        m: mass of wet air
        rhov: mass density of vapor
        rho: mass density of wet air
        P: atmospheric pressure (hPa)
    """
    q = (622 * e) / (P - 0.378 * e)   
    return q

def mixing_ratio(e, P):
    """
    Input:
    ------
    e(hPa): 
        vapor pressure
    P(hPa):
        atmospheric pressure
    Output:
    -------
    r(g/kg):
        mixing ratio, the ratio of vapor mass to dry air mass
    math:
    -----
    r = mv / md = rhov / rhod = (e / (Rv * T)) / (Pd / (Rd * T)) = (Rd * e) / (Rv * (P - e)) 
      = 622 * e / (p - e) g/Kg = 0.622 * e / (p - e) g/g
    where
        mv: mass of vapor
        md: mass of dry air
        rhov: mass density of vapor
        rhod: mass density of dry air
        Rv: water vapor gas constant 461.5(J Kg^-1 K^-1)
        Rd: dry air gas constant 286.9(J Kg^-1 K^-1)
    """
    r = (622 * e) / (P - e)
    return r

def virtual_temperature(T, solution, r = None, q = None, Td = None, P = None):
    """
    Input:
    ------
    T(degC):
        temperature
    r(kg/kg): optional
        mixing ratio, the ratio of vapor mass to dry air mass
    q(kg/kg): optional
        Specific humidity
    Td(degC): optional
        Dewpoint temperature
    P(hPa): optional
        P denotes the atmospheric pressure at the point of measurement (hPa)
    Output:
    Tv(degC):
        virtual temperature
    ------        
    Note:
    -----
    Solution 1: https://glossary.ametsoc.org/wiki/Virtual_temperature#:~:text=(Also%20called%20density%20temperature.),and%20water%20vapor%2C%20%E2%89%88%200.622.
        use T and r
    Solution 2:
        use T and q
    Solution 3: https://www.weather.gov/media/epz/wxcalc/virtualTemperature.pdf
        use T, Td, and P
    """
    if solution == 1:
        assert r != None, "mixing ratio r(Kg/Kg) must be given for solution 1" 
        Tv = T * (1 + r / 0.622) / (1 + r)
    elif solution == 2:
        assert q != None, "specific humidity q(Kg/Kg) must be given for solution 2" 
        Tv=(1 + 0.608 * q)* T
    elif solution == 3:
        assert (Td != None) and (P != None), "dewpoint temperature Td(degC) and station pressure P(hPa) must be given for solution 3" 
        Tv = (T + 273.15) / (1 - 0.379 * (6.11 * 10**((7.5 * Td) / (237.7 + Td)) / P))
        Tv = Tv - 273.15      
    else:
        raise(Exception("solution must be given as one of 1, 2, and 3"))
        
    return Tv

# dry air density
def dry_air_density(T, P, solution = 1):
    """
    Parameters:
    -----------
    T(degC): 
        Air temperature
    P(hPa): 
        Pressure
    solution: int
        Choice of formula
    Returns:
    --------
    rho_dry(g/m3): 
        Dry air density
    math:
    -----
    rho = M/V
        where M is mass (kg), V is volume (m3)
    according to gas equation:
    rho = rho0 * (T0 * P) / (P0 * T) = 3.48
        where subscript 0 is stadard case, and P0 = 101.325 KPa, T0 = 273.15 K, rho0 (dry air density) is 1.293 kg/m3
    """
    # solution 1:
    if solution == 1:
        rho0 = 1.293e3
        rho_dry = rho0 * 273.15 / (T + 273.15) * P / 1013.25
    # solution 2:
    else: 
        P = P / 10 # unit: KPa
        rho_dry = 3.48 * P / (T + 273.15) # Kg/m3
        rho_dry = rho_dry * 1000
    return rho_dry

def wet_air_density(T, P, es, MH2O = 1.8015268e-2, Md = 28.96546e-3):
    """
    Parameters:
    -----------
    T(K): 
        Air temperature
    P(Pa):
        Pressure
    es(Pa):
        Saturation_vapor_pressure
    MH2O(kg/mol): optional
        Water molecular weight
    Md(kg/mol): optional
        Dry air molecular weight
    Returns:
    --------
    rho_wet(g/m3): wet air density
    theory:
    -------
    Ideal gas law and Dalton partial pressure
    This method has large bias.
    """
    rho_wet = (MH2O * es + (P - es) * Md) / (8.314 * T) # Kg/m3
    rho_wet = rho_wet * 1000
    return rho_wet    

def wet_air_density2(T, P, RH, es):
    """
    Parameters:
    -----------
    T(degC): 
        Air temperature
    P(hPa):
        Pressure
    RH: 
        Relative humidity in the range [0 - 1]
    es(hPa):
        Saturation_vapor_pressure
    Returns:
    --------
    rho_wet(g/m3): wet air density
    theory:
    -------
    Ideal gas law and Dalton partial pressure
    This method has large bias.
    """
    P = P / 10 # unit: KPa
    rho_wet = 3.48 * P/(T + 273.15) * (1- 0.378 * RH * es / P) # Kg/m3
    rho_wet = rho_wet * 1000
    return rho_wet

# -------------------------------------------------------------------
# Advanced equations
def vapor_pressure(RH, T):
    """
    Vapor pressure at temperature T is the saturation vapor pressure at the corresponding dewpoint temperature
    Input:
    ------
    RH:
        Relative humidity in the range [0 - 1]
    T(degC):
        temperature
    Output:
    -------
    e(hPa): vapor pressure at temperature T
    """
    Td = relative_humidity2dewpoint(T, RH)
    e = saturation_vapor_pressure(Td)
    return e 


# -------------------------------------------------------------------
# Eddy Covariance meteo equations
def virtual_temperature_from_absolute_real_temperature(T, e, P):
    """
    Input:
    ------
    T(K):
        absolute real temperature
    e(hPa): 
        vapor pressure
    P(hPa):
        atmospheric pressure
    Output:
    -------
    virtual temperature(K)
    """
    virtual_temperature = T * (1 + 0.38 * e / P)
    return virtual_temperature

def absolute_real_temperature_from_virtual_temperature(T, e, P):
    """
    Input:
    ------
    T(K):
        virtual temperature
    e(hPa): 
        vapor pressure
    P(hPa):
        atmospheric pressure
    Output:
    -------
    absolute real temperature(K)
    """
    absolute_real_temperature = T / (1 + 0.38 * e / P)
    return absolute_real_temperature

def sonic_temperature_from_absolute_real_temperature(T, e, P):
    """
    Input:
    ------
    T(K):
        absolute real temperature
    e(hPa): 
        vapor pressure
    P(hPa):
        atmospheric pressure
    Output:
    -------
    sonic temperature(K)
    """
    sonic_temperature = T * (1 + 0.32 * e / P)
    return sonic_temperature

def absolute_real_temperature_from_sonic_temperature(T, e, P):
    """
    Input:
    ------
    T(K):
        absolute real temperature
    e(hPa): 
        vapor pressure
    P(hPa):
        atmospheric pressure
    Output:
    -------
    sonic temperature(K)    
    """
    absolute_real_temperature = T / (1 + 0.32 * e / P)
    return absolute_real_temperature


# -*- coding: utf-8 -*-
"""
Created on Tue Jan 05 16:42:00 2021
V1 on Thu Jan 07 09:07:00 2021

@author: sz394@exeter.ac.uk
"""

import warnings
import numpy as np

def crosscorr(Uz, scalar, lag = 0):
    """Lag-N cross correlation. 
    Parameters
    ----------
    Uz, scaler : pandas.Series objects of equal length, Uz is vertial wind, scaler is CO2, H2O etc.
    lag : int, default 0

    Returns
    ----------
    corr : float
    
    Note:
    -----
    Pearson correlation (method 1):
        keyword can be 'pearson', 'kendall', 'spearman'
        https://stackoverflow.com/questions/33171413/cross-correlation-time-lag-correlation-with-pandas
    Covariance correlation (methhod 2):
        corr = cov(x, y) / (std(x) * std(y))
        https://study.com/academy/lesson/covariance-correlation-equations-examples.html#:~:text=The%20correlation%20coefficient%20is%20represented,standard%20deviations%20of%20each%20variable.&text=A%20correlation%20coefficient%20uses%20the,takes%20it%20one%20step%20further.
    
    Math:
    -----
    Scalar(e.g. CO2) is the lagged variable due to the tube wall effect etc.
    for example(df):
        Uz  |CO2
        ---------
        8:00|7:00
        9:00|8:00
    so the scalar = scalar.shift(-lag) --> Note the "-" sign (lag is positive)
        Uz  |CO2
        ---------
        8:00|8:00
        9:00|NaN    
    cov = SUM(
        (x - mean(x)) * (y - mean(y))
    ) / (N-1)
    """
    scalar = scalar.shift(-lag)
    corr = Uz.cov(scalar) / (Uz.std() * scalar.std())
    # # or (not recommended by its similar result yet lower speed):
    # corr = Uz.corr(scalar, method = "spearman")
    return corr

def remove_lag(df, Uz, scalar, window = [-50, 200], hz = 10, thre_q = 0.75, fco2 = 1):
    """
    Parameters
    ----------
    df : Pandas DataFrame
        EC data containing vertical velocity(UZ) and scalar(e.g. CO2)
    velocity: str
        name of vertical wind in df
    scalar: str
        name of CO2, H2O etc in df
    window: list, default is [-50, 200], because normally scalar signal is lagged behind wind
        range for shifting, the unit is EC freq, 100 ms for 10Hz systems.
    hz: int, default is 10
        EC system frequency
    thre_q: percentile to determine the sign of covariance correlation
    
    Returns
    ---------
    pandas.Series: lag removed series
    float: delay seconds
    """
    # window = range(window[0], window[1])
    coefs = [
      crosscorr(df[Uz], df[scalar], lag = l) for l in range(*window)
    ]
    
    coefs = np.array(coefs)
    window = np.arange(*window)

    # ----------------------------------------------
    # decide choose max or min coef
    timestamp = (df.index[0] + (df.index[-1] - df.index[0]) / 2).replace(microsecond=0)
    hour = timestamp.hour
    choice = -1 # -1 for min, 1 for max 
    if (hour > 8) and (hour < 16): # day, cov < 0
        choice = -1 # n_shift = window[np.argmin(coefs)]
    elif (hour < 5) or (hour > 21): # night, cov > 0
        choice = 1 # n_shift = window[np.argmax(coefs)]
    else:
        thre_sun = np.quantile(coefs, thre_q)
        if thre_sun > 0:
            choice = 1 # n_shift = window[np.argmax(coefs)]
        else:
            choice = -1 # n_shift = window[np.argmin(coefs)]
    if choice * fco2 == -1:
        n_shift = window[np.argmin(coefs)]
    else:
        n_shift = window[np.argmax(coefs)]
    # ----------------------------------------------
    # what if n_shift too far
    
    df[scalar] = df[scalar].shift(-n_shift)
    lag_sec = n_shift / hz
    return df, lag_sec


from .edire import EdiReTbx
from .utils import unit2unit, fetch_sort_path, group_day
from .df2tob1 import dataframe2tob1

__all__ = ["EdiReTbx", "unit2unit", "dataframe2tob1", "fetch_sort_path", "group_day"]


import numpy as np
import pandas as pd
from .CO2 import center_rolling


def stats4qc(ts):
    # statistics
    q1 = ts.quantile(0.25)
    q3 = ts.quantile(0.75)
    iqr = q3 - q1
    low = q1 - 1.5*iqr
    high = q3 + 1.5*iqr
    return [low, high]

def iqr_qc(ts, low, high):
    return ts[(ts < low) | (ts > high)].index

def glbdec(ts):
    ts = ts.copy()
    tsr = ((ts / center_rolling(ts, 601)).abs() - 1) * 100
    tsr = tsr.abs().resample("600S").mean()
    tsr = tsr.resample("100ms").interpolate()
    tsr = tsr[ts.index]

    ts[iqr_qc(tsr, *stats4qc(tsr))] = np.nan
    tsr = ((ts / center_rolling(ts, 101)).abs() - 1) * 100
    ts[tsr.abs() > 2] = np.nan
    return ts

def despike(ts, threshold = 2):
    ts = ts.copy()
    tsr = ((ts / center_rolling(ts, 101)).abs() - 1) * 100
    ts[tsr.abs() > threshold] = np.nan
    return ts

def detrend(ts, order = 3):
    ts = ts.copy()
    coefs = np.polyfit(np.arange(len(ts)), ts.interpolate().bfill().values, order)
    ffit = np.polyval(coefs, np.arange(len(ts)))
    ffit = pd.Series(ffit, index = ts.index)
    return ts - ffit

def deplatform(ts, order = 2):
    ts = ts.copy()
    tsd = detrend(ts, order = order)
    # print(tsd)
    # tsr = ((tsd / center_rolling(tsd, 101, method = 'mean')).abs() - 1) * 100
    # tsc = ts.copy()
    ts[iqr_qc(tsd, *stats4qc(tsd))] = np.nan
    return ts


# -*- coding: utf-8 -*-
import matplotlib.pyplot as plt
from scipy import signal
"""
Created on Thu Jan 07 17:13:00 2021

@author: sz394@exeter.ac.uk
"""

class Spec():
    def __init__(self, hz = 10):
        self.hz = hz
        
    def spectra(self, scalar):
        """
        calculate scalar spectra
        Parameters:
        -----------
        scalar: pandas.Series
        Returns:
        --------
        freq: numpy.ndarray frequencies
        psd(spectra): numpy.ndarray power spectral density
        """
        scalar = scalar.interpolate(method = "linear")

        # method 1: welch
        freqs, psd = signal.welch(
            scalar,
            fs = 1/self.hz, # # sample rate
            window = "hamming",   # apply a Hanning window before taking the DFT
            detrend = "constant", # detrend scalar by subtracting the mean
            scaling = "spectrum"
        )

        # # method 2: csd (same to welch)
        # freqs, psd = signal.csd(
        #     scalar, scalar, 
        #     fs = 1/self.hz, # # sample rate,
        #     window = "hamming",   # apply a Hanning window before taking the DFT
        #     detrend = "constant", # detrend scalar by subtracting the mean
        #     scaling = "spectrum" # Selects between computing the power spectral density ('density') where Pxx has units of V**2/Hz and computing the power spectrum ('spectrum') where Pxx has units of V**2, if x is measured in V and fs is measured in Hz.
        # )
        return freqs, psd

    def cospectra(self, wind, scalar):
        """
        calculate scalar cospectra
        Parameters:
        -----------
        wind: pandas.Series
        scalar: pandas.Series
        Returns:
        --------
        freq: numpy.ndarray frequencies
        csd(cospectra): numpy.ndarray power spectral density
        """
        scalar = scalar.interpolate(method = "linear")
        wind = wind.interpolate(method = "linear")
        
        freqs, csd = signal.csd(
            scalar,
            wind, 
            fs = 1/self.hz, # # sample rate,
            window = "hamming",   # apply a Hanning window before taking the DFT
            detrend = "constant", # detrend scalar by subtracting the mean
            scaling = "spectrum" # Selects between computing the power spectral density ('density') where Pxx has units of V**2/Hz and computing the power spectrum ('spectrum') where Pxx has units of V**2, if x is measured in V and fs is measured in Hz.
        )
        csd = csd.real
        # csd = np.abs(csd)
        return freqs, csd

    def plot(self, freqs, density, figsize = (5, 4), title = "", xlabel = "Frequency", ylabel = "Power", savefile = ""):
        """
        plot spectra and cospectra
        Parameters:
        --------
        freq: numpy.ndarray frequencies
        csd(cospectra): numpy.ndarray power spectral density
        Returns:
        --------
        None
        """
        fig, ax = plt.subplots(figsize = figsize)
        ax.semilogx(freqs, density)
        # # equivalent to
        # ax.plot(freqs, density)
        # ax.set_xscale('log')
        ax.set_title(title, fontsize = 14)
        ax.set_xlabel(xlabel, fontsize = 12)
        ax.set_ylabel(ylabel, fontsize = 12)
        ax.tick_params(direction = "in", which = "both", labelsize = 10)
        plt.tight_layout()
        if savefile:
            fig.savefig(savefile, dpi = 300, bbox_inches = "tight")
        plt.close('all')


#!/usr/bin/env python
#-*- coding:utf-8 -*-

#############################################
# File Name: setup.py
# Author: Songyan Zhu
# Mail: sz394@exeter.ac.uk
# Created Time:  2021-01-05
#############################################


from setuptools import setup, find_packages

setup(
	name = "eddyflux",
	version = "0.0.7",
	keywords = ("eddy covariance processing", "flux"),
	description = "left blank",
	long_description = "left blank",
	license = "MIT Licence",

	url = "",
	author = "Songyan Zhu",
	author_email = "sz394@exeter.ac.uk",

	packages = find_packages(),
	include_package_data = True,
	platforms = "any",
	install_requires=[

	]
)


# author: Songyan Zhu
# concat: sz394@exeter.ac.uk
# Spectra/cospectra analysis converted from matalb codes of
# Dr. Robert Clement and Dr. Timothly Hill

from .specutils import SplitFFT, SplitCSV, numlist2arr, fetch
from .spectra_analysis import SelectStructure, CalcLambda, TubeAtten, FreqResp2, VectorModelSpectra, AvgSpectra

__all__ = [
    'SplitFFT', 'SplitCSV', 'numlist2arr', 'fetch',
    'SelectStructure', 'CalcLambda', 'TubeAtten', 'FreqResp2', 'VectorModelSpectra', 'AvgSpectra'
    ]


# -*- coding: utf-8 -*-
"""
Created on Fri Nov 6 14:44:36 2020
Modified on Wed Dec 30 15:42:37 2020

@author: sz394@exeter.ac.uk
"""
import numpy as np
import pandas as pd
from pathlib import Path

class EdiReTbx():
    def __init__(self):
        pass
    
    def make_input_list(self, search_directory, output_file, loc = 6, screen = 0, remove_imcomplete = True, user_file_size = None):
        """
        Created on Fri Nov 6 14:44:36 2020
        Modified on Dec 30 2020: make the functions into a class.
        Search and save
        Parameters
        ----------
        search_directory : Path
            Search directory.
        output_file : Path
            Save directory.
        loc : int, optional
            The location of year in splited string by _, count starting from 0. The default is 6.
        screen : int, optional
            0: no screen output, 1: print error info, 2: print everything. The default is 0.
        remove_imcomplete : boolean, optional
            Ignore incomplete files. The default is True.
        user_file_size : float, optional
            File size threshold. The default is None.

        Returns
        -------
        None.
        
        Examples
        --------
        >>> search_directory = r"C:/Users/sz394/OneDrive - University of Exeter/Workspace/low_cost_eddy_covariance/Data/2020_09_26/Converted_200105"
        >>> output_file = r"C:/Users/sz394/OneDrive - University of Exeter/Workspace/low_cost_eddy_covariance/Data/2020_09_26/EdiRe_200105/input/input_list.txt"
        >>> ertbx = EdiReTbx()
        >>> ertbx.make_input_list(search_directory, output_file)
        """
        df = self.__search(search_directory, loc, screen, remove_imcomplete, user_file_size)
        self.__write2file(df, output_file)
        

    def __search(self, search_directory, loc, screen, remove_imcomplete, user_file_size):
        """
        Search tob file and return directories as a pandas dataframe
        Parameters
        ----------
        search_directory : Path
            Folder directory.
        loc : int
            The location of year in splited string by _, count starting from 0.
        screen : int
            0: no screen output, 1: print error info, 2: print everything.
        remove_imcomplete : boolean
            Ignore incomplete files.
        user_file_size : float
            File size threshold.

        Returns
        -------
        df : DataFrame
            Contains searched directories, columns are "directory", "year", "month", "day", "time", "file size".

        """
        #---------------------------------------------------------------------------------------------------------------------------------------
        paths = list(Path(search_directory).glob("*.dat"))

        records = []
        for pos, p in enumerate(paths):
            kwds = p.stem.split("_")
            # # example
            # kwds = "TOB1_13192_roth_200102_ts_9_2020_08_27_2300".split("_")

            if (len(kwds) > loc + 3):
                year = kwds[loc]
                month = kwds[loc + 1]
                day = kwds[loc + 2]
                minsec = kwds[loc + 3]

                if screen == 2:
                    print(year, month, day)
                if (len(year) == 4) & (len(month) == 2) & (len(day) == 2) & (len(minsec) == 4):
                    records.append([p, year, month, day, minsec, p.stat().st_size])
                else:
                    if screen:
                        print(
                            f"year length: {len(year)}, month length: {len(month)}, day length: {len(day)}"
                        )
                    else:
                        pass
            else:
                pass

        df = pd.DataFrame(records, columns = ["directory", "year", "month", "day", "time", "file size"])
        df = df.sort_values(["year", "month", "day", "time"], ascending=[True, True, True, True])

        # remove imcomplete files:
        if remove_imcomplete:
            if not user_file_size:
                max_size = df["file size"].max()
                df = df[df["file size"] == max_size]
            else:
                df = df[df["file size"] >= user_file_size]

        print("Filtering is finished...")
        return df

    def __write2file(self, df, output_file):
        with open(output_file, 'w') as f:
            f.writelines("%s\n" % placeholder.as_posix() for placeholder in df["directory"])
        print("Writting is finished...")


# -*- coding: utf-8 -*-
"""
Created on Sat Jan  2 08:25:53 2021

@author: University of Manitoba - CEOS
@website: https://github.com/UofM-CEOS/flux_capacitor
"""

import numpy as np
from collections import namedtuple
PlanarFitCoefs = namedtuple("PlanarFitCoefs",
                            ["k_vct", "tilt_coefs", "phi", "theta"])

RotatedVectors = namedtuple("RotatedVectors", ["rotated", "phi_theta"])

# Valid 3-D wind rotation methods
_VECTOR_ROTATION_METHODS = {"DR", "TR", "PF"}

def planarfit(vectors):
    """Calculate planar fit coefficients for coordinate rotation
    See Handbook of Micrometeorology (Lee et al. 2004).  Ported from
    getPlanarFitCoeffs.m from Patric Sturm <pasturm@ethz.ch>.
    Parameters
    ----------
    vectors : numpy.ndarray
        A 2-D (Nx3) array with `x`, `y`, and `z` vectors, expressed in a
        right-handed coordinate system.  These vectors may correspond to
        `u`, `v`, and `w` wind speed vectors, or inertial acceleration
        components.
    Returns
    -------
    PlanarFitCoefs : namedtuple
        namedtuple with (index and name in brackets):
        numpy.ndarray [0, `k_vct`]
            1-D array (1x3) unit vector parallel to the new z-axis.
        numpy.ndarray [1, `tilt_coefs`]
            1-D array (1x3) Tilt coefficients `b0`, `b1`, `b2`.
        numpy.float [2, `phi`]
            Scalar representing roll angle :math:`\\phi`.
        numpy.float [3, `theta`]
            Scalar representing pitch angle :math:`\\theta`.
    """
    vct_u = vectors[:, 0]
    vct_v = vectors[:, 1]
    vct_w = vectors[:, 2]
    vct_nrows = vectors.shape[0]
    sum_u = sum(vct_u)
    sum_v = sum(vct_v)
    sum_w = sum(vct_w)
    dot_uv = np.dot(vct_u, vct_v)
    dot_uw = np.dot(vct_u, vct_w)
    dot_vw = np.dot(vct_v, vct_w)
    dot_u2 = np.dot(vct_u, vct_u)
    dot_v2 = np.dot(vct_v, vct_v)
    H_arr = np.array([[vct_nrows, sum_u, sum_v],
                      [sum_u, dot_u2, dot_uv],
                      [sum_v, dot_uv, dot_v2]])
    g_arr = np.array([sum_w, dot_uw, dot_vw])
    tilt_coef = np.linalg.solve(H_arr, g_arr)
    # Estimated \phi (roll) and \theta (pitch) tilt angles
    phi_denom = np.sqrt(1 + (tilt_coef[2] ** 2))
    phi_sin = tilt_coef[2] / phi_denom
    phi_cos = 1 / phi_denom
    phi = np.arctan2(phi_sin, phi_cos)
    theta_denom = np.sqrt(1 + (tilt_coef[1] ** 2) + (tilt_coef[2] ** 2))
    theta_sin = -tilt_coef[1] / theta_denom
    theta_cos = np.sqrt((tilt_coef[2] ** 2) + 1) / theta_denom
    theta = np.arctan2(theta_sin, theta_cos)
    # Determine unit vector parallel to new z-axis
    k_2 = 1 / np.sqrt(1 + tilt_coef[1] ** 2 + tilt_coef[2] ** 2)
    k_0 = -tilt_coef[1] * k_2
    k_1 = -tilt_coef[2] * k_2
    k_vct = np.array([k_0, k_1, k_2])
    return PlanarFitCoefs(k_vct, tilt_coef, phi, theta)

def rotate_wind3d(wind3D, method="PF", **kwargs):
    """Transform 3D wind vectors to reference mean streamline coordinate system
    Use double rotation, triple rotation, or planar fit methods (Wilczak et
    al. 2001; Handbook of Micrometeorology).
    This is a general coordinate rotation tool, so can handle inputs such
    as wind speed and acceleration from inertial measurement units.
    Parameters
    ----------
    wind3D : numpy.ndarray
        A 2-D (Nx3) array with `x`, `y`, and `z` vector components,
        expressed in a right-handed coordinate system.  These may represent
        `u`, `v`, and `w` wind speed vectors, or inertial acceleration.
    method : {"DR", "TR", "PF"}, optional
        One of: "DR", "TR", "PF" for double rotation, triple rotation, or
        planar fit.
    k_vector : numpy.ndarray, optional
        1-D array (1x3) unit vector parallel to the new z-axis, when
        "method" is "PF" (planar fit).  If not supplied, then it is
        calculated.
    Returns
    -------
    RotatedVectors : namedtuple
        namedtuple with (index, name in brackets):
        numpy.ndarray [0, `rotated`]
            2-D array (Nx3) Array with rotated vectors
        numpy.ndarray [1, `phi_theta`]
            1-D array (1x2) :math:`\\phi` and :math:`\\theta` rotation
            angles.  The former is the estimated angle between the vertical
            unit coordinate vector in the rotated frame and the vertical
            unit vector in the measured uv plane, while the latter is wind
            direction in the measured uv plane.  Note these are *not* roll
            and pitch angles of the measurement coordinate frame relative
            to the reference frame.
    """
    if method not in _VECTOR_ROTATION_METHODS:
        msg = "method must be one of "
        raise ValueError(msg + ', '.join("\"{}\"".format(m) for m in
                                         _VECTOR_ROTATION_METHODS))

    if method == "PF":
        if "k_vector" in kwargs:
            k_vct = kwargs.get("k_vector")
        else:
            pfit = planarfit(wind3D)
            k_vct = pfit.k_vct
        j_vct = np.cross(k_vct, np.mean(wind3D, 0))
        j_vct = j_vct / np.sqrt(np.sum(j_vct ** 2))
        i_vct = np.cross(j_vct, k_vct)
        vcts_mat = np.column_stack((i_vct, j_vct, k_vct))
        vcts_new = np.dot(wind3D, vcts_mat)
        phi = np.arccos(np.dot(k_vct, np.array([0, 0, 1])))
        theta = np.arctan2(np.mean(wind3D[:, 1], 0),
                           np.mean(wind3D[:, 0], 0))
    else:
        # First rotation to set mean v to 0
        theta = np.arctan2(np.mean(wind3D[:, 1]),
                           np.mean(wind3D[:, 0]))
        rot1 = np.array([[np.cos(theta), -np.sin(theta), 0],
                         [np.sin(theta), np.cos(theta), 0],
                         [0, 0, 1]])
        vcts1 = np.dot(wind3D, rot1)
        # Second rotation to set mean w to 0
        phi = np.arctan2(np.mean(vcts1[:, 2]),
                         np.mean(vcts1[:, 0]))
        rot2 = np.array([[np.cos(phi), 0, -np.sin(phi)],
                         [0, 1, 0],
                         [np.sin(phi), 0, np.cos(phi)]])
        vcts_new = np.dot(vcts1, rot2)
        # Third rotation to set mean vw to 0
        if method == "TR":
            mean_vw = np.mean(vcts_new[:, 1] * vcts_new[:, 2])
            psi = 0.5 * np.arctan2(2 * mean_vw,
                                   np.mean(vcts_new[:, 1] ** 2) -
                                   np.mean(vcts_new[:, 2] ** 2))
            rot3 = np.array([[1, 0, 0],
                             [0, np.cos(psi), -np.sin(psi)],
                             [0, np.sin(psi), np.cos(psi)]])
            vcts_new = np.dot(vcts_new, rot3)

    return RotatedVectors(vcts_new, np.array([phi, theta]))

def rotate_wind(vectors, method = "PF", return_all = False):
    """
    Wrapper of rotate_wind3d
    Parameters
    ----------
    wind3D : numpy.ndarray
        A 2-D (Nx3) array with `x`, `y`, and `z` vector components,
        expressed in a right-handed coordinate system.  These may represent
        `u`, `v`, and `w` wind speed vectors, or inertial acceleration.
    method : {"DR", "TR", "PF"}, optional
        One of: "DR", "TR", "PF" for double rotation, triple rotation, or
        planar fit.
    return_all: boolean
        Return RotatedVectors if True, rotated array only otherwise. Default: False
    Returns
    -------
    rotated_vectors:
        numpy.ndarray [0, `rotated`]
            2-D array (Nx3) Array with rotated vectors
    or return_all is True
    rotated_tuple(RotatedVectors) : namedtuple
        namedtuple with (index, name in brackets):
        numpy.ndarray [0, `rotated`]
            2-D array (Nx3) Array with rotated vectors
        numpy.ndarray [1, `phi_theta`]
            1-D array (1x2) :math:`\\phi` and :math:`\\theta` rotation
            angles.  The former is the estimated angle between the vertical
            unit coordinate vector in the rotated frame and the vertical
            unit vector in the measured uv plane, while the latter is wind
            direction in the measured uv plane.  Note these are *not* roll
            and pitch angles of the measurement coordinate frame relative
            to the reference frame.
    """
    rotated_tuple = rotate_wind3d(vectors, method = method)
    if return_all:
        return rotated_tuple
    else:
        rotated_vectors = rotated_tuple[0]
        return rotated_vectors


# -*- coding: utf-8 -*-
import numpy as np
from scipy.stats import zscore
from scipy import interpolate as itpl
from collections import namedtuple
from itertools import groupby
"""
Created on Sat Jan  2 09:14:14 2021

@author: University of Manitoba - CEOS
@website: https://github.com/UofM-CEOS/flux_capacitor
"""
VickersMahrt = namedtuple("VickersMahrt",
                          ["x", "nspikes", "ntrends", "kclass"])

def window_indices(idxs, width, step=None):
    """List of sliding window indices across an index vector
    Parameters
    ----------
    idx : numpy.ndarray
        A 1-D index vector.
    width : int
        Window width.
    step : int, optional
        Step size for sliding windows.
    Returns
    -------
    list
        List of tuples, each with the indices for a window.
    """
    return zip(*(idxs[i::step] for i in range(width)))

def get_VickersMahrt(x, zscore_thr, nrep_thr):
    """Vickers Mahrt computations in a window
    Parameters
    ----------
    x : numpy.ndarray
        A 1-D signal vectors to be despiked.
    zscore_thr : float
        The zscore beyond which an observation is considered to be an
        outlier.
    nrep_thr : int
        The maximum number of consecutive outliers that should occur for a
        spike to be detected.
    Returns
    -------
    VickersMahrt : tuple
        Tuple with (index, name in brackets):
        numpy.ndarray [0, `x`]
            1-D array with interpolated input.
        numpy.int [1, `nspikes`]
            Number of spikes detected.
        numpy.int [2, `ntrends`]
            Number of outlier trends detected.
        numpy.ndarray [3, `kclass`]
            1-D array of the same size as input, indicating the
            classification `k` for each measurement. k=0: measurement
            within plausibility range, k=[-1 or 1]: measurement outside
            plausibility range, abs(k) > 1: measurement is part of an
            outlier trend.
    """
    z = zscore(x)
    # Discern between outliers above and below the threshold
    isout_hi, isout_lo = (z > zscore_thr), (z < -zscore_thr)
    n_outs = sum(isout_hi) + sum(isout_lo)
    # Set categorical x: 0 (ok), 1 (upper outlier), -1 (lower outlier)
    xcat = np.zeros(x.shape, dtype=np.int)
    xcat[isout_hi] = 1
    xcat[isout_lo] = -1
    if n_outs > 0:
        # Create tuples for each sequence indicating whether it's outliers
        # and its length.
        grps = [(val, len(list(seq))) for val, seq in groupby(xcat)]
        vals = np.array([k[0] for k in grps])
        lens = np.array([k[1] for k in grps])
        is_spike = (vals != 0) & (lens <= nrep_thr)
        nspikes = sum(is_spike)
        # We tally trends as well
        is_trend = (vals != 0) & (lens > nrep_thr)
        ntrends = sum(is_trend)
        # If we have trends, loop through each one, knowing the length of
        # the spike and where we are along the input series.
        if ntrends > 0:
            trends = zip(vals[is_trend], lens[is_trend],
                         np.cumsum(lens)[is_trend])
            for i in trends:
                # Double the categorical value for trends and consider
                # these OK for interpolation. So abs(xcat) > 1 are trends.
                xcat[(i[2] - i[1]):i[2]] = i[0] * 2
        # Now we are left with true outliers to interpolate
        x_new = x.copy()
        xidx = np.arange(len(x))             # simple index along x
        isok = (xcat == 0) | (abs(xcat) > 1)  # ok if 0 or trend
        s = itpl.InterpolatedUnivariateSpline(xidx[isok],
                                              x[isok], k=1)
        x_itpl = s(xidx[~ isok])
        x_new[~ isok] = x_itpl
        return VickersMahrt(x_new, nspikes, ntrends, xcat)
    else:
        return VickersMahrt(x, 0, 0, xcat)


def despike_VickersMahrt(x, width, zscore_thr, nreps, step=None,
                         nrep_thr=None, interp_nan=True):
    """Vickers and Mahrt (1997) signal despiking procedure
    The interpolating function is created by the
    InterpolatedUnivariateSpline function from the scipy package, and uses
    a single knot to approximate a simple linear interpolation, so as to
    keep the original signal as untouched as possible.
    Parameters
    ----------
    x : numpy.ndarray
        A 1-D signal vectors to be despiked.
    width : int
        Window width.
    step : int, optional
        Step size for sliding windows.  Default is one-half window width.
    zscore_thr : float
        The zscore beyond which an observation is considered to be an
        outlier.  Default is zero.
    nrep_thr : int, optional
        The maximum number of consecutive outliers that should occur for a
        spike to be detected.  Default: 3.
    nreps: int, optional
        How many times to run the procedure.  Default is zero.
    interp.nan : bool, optional
        Whether missing values should be interpolated.  Interpolated values
        are computed after despiking.  Default is True.
    Returns
    -------
    VickersMahrt : tuple
        Tuple with (index, name in brackets):
        numpy.ndarray [0, `x`]
            1-D array with despiked input.
        numpy.int [1, `nspikes`]
            Number of spikes detected.
        numpy.int [2, `ntrends`]
            Number of outlier trends detected.
        numpy.int [3, `kclass`]
            Number of iterations performed.
    """
    if step is None:            # set default step as
        # step = width / 2        # one-half window size
        # sz394: step should be integer
        step = width // 2        # one-half window size
    if nrep_thr is None:
        nrep_thr = 3
    nspikes, ntrends = 0, 0
    xout = x.copy()
    # Following EddyUH implementation, fill missing values with nearest
    # value for the purpose of spike and trend detection.  This ensures
    # that we can always calculate zscores.
    is_missing = np.isnan(np.array(xout))  # need to coerce to np array
    xidx = np.arange(len(xout))            # simple index along x
    f_itpl = itpl.interp1d(xidx[~ is_missing], xout[~ is_missing],
                           kind="nearest", fill_value="extrapolate")
    x_nonan = f_itpl(xidx)
    # Get a series of tuples with indices for each window
    idxl = window_indices(range(len(x)), width, step)
    nloops = 0
    while nloops < nreps:
        nspikes_loop = 0
        for w in idxl:
            winidx = [i for i in w]  # indices of current window
            xwin = x_nonan[winidx]   # values for the current window
            xnew, nsp, ntr, xmask = get_VickersMahrt(xwin, zscore_thr,
                                                     nrep_thr)
            nspikes_loop += nsp
            ntrends += ntr
            x_nonan[winidx] = xnew
        nloops += 1
        # Increase zscore_thr by 0.3, instead of 0.1 as in V&M (1997),
        # following EddyUH implementation
        zscore_thr += 0.3
        if nspikes_loop > 0:
            nspikes += nspikes_loop
        else:                   # Stop if we haven't found any new spikes
            break
    # Interpolate through missing values, if requested (default).
    nmissing = np.count_nonzero(is_missing)
    if (nmissing > 0) and interp_nan:
        s = itpl.InterpolatedUnivariateSpline(xidx[~ is_missing],
                                              xout[~ is_missing], k=1)
        x_itpl = s(xidx[is_missing])
        xout[is_missing] = x_itpl

    return VickersMahrt(xout, nspikes, ntrends, nloops)


# despike
def despikeVM(x, width, return_all = False, zscore_thr = 0, nreps = 0, step=None, nrep_thr=None, interp_nan=True):
    """
    Wrapper of Vickers and Mahrt despike
    Parameters
    ----------
    x : numpy.ndarray
        A 1-D signal vectors to be despiked.
    width : int
        Window width.
    return_all: boolean
        Return VickersMahrt if True, despiked array only otherwise. Default: False
    zscore_thr : float
        The zscore beyond which an observation is considered to be an
        outlier.  Default is zero.
    nreps: int, optional
        How many times to run the procedure.  Default is zero.
    step : int, optional
        Step size for sliding windows.  Default is one-half window width.
    nrep_thr : int, optional
        The maximum number of consecutive outliers that should occur for a
        spike to be detected.  Default: 3.
    interp.nan : bool, optional
        Whether missing values should be interpolated.  Interpolated values
        are computed after despiking.  Default is True.
    Returns
    -------
    despiked_array:
        numpy.ndarray [0, `x`]
            1-D array with despiked input.
    or return_all is True
    despiked_tuple(VickersMahrt) : tuple
        Tuple with (index, name in brackets):
        numpy.ndarray [0, `x`]
            1-D array with despiked input.
        numpy.int [1, `nspikes`]
            Number of spikes detected.
        numpy.int [2, `ntrends`]
            Number of outlier trends detected.
        numpy.int [3, `kclass`]
            Number of iterations performed.
    """
    despiked_tuple = despike_VickersMahrt(
        x, width, zscore_thr, nreps, 
        step = step, nrep_thr = nrep_thr, interp_nan = interp_nan
    )
    if return_all:
        return despiked_tuple
    else:
        despiked_array = despiked_tuple[0]
        return despiked_array